#!/usr/bin/env python
"""
A script setting the metadata of segmentation models.
"""
import shutil

import click


@click.command()
@click.option('-b', '--bounding-region', multiple=True, help='Sets region identifiers which bound line bounding polygons')
@click.option('--topline/--baseline', default=False, help='Sets model line type to baseline or topline')
@click.option('--pad', show_default=True, type=(int, int), default=(0, 0),
              help='Padding (left/right, top/bottom) around the page image')
@click.argument('model', nargs=1, type=click.Path(exists=True))
def cli(bounding_region, topline, pad, model):
    """
    A script setting the metadata of segmentation models.
    """
    from kraken.lib import vgsl

    net = vgsl.TorchVGSLModel.load_model(model)
    if net.model_type != 'segmentation':
        print('Model is not a segmentation model.')
        return

    print('detectable line and region types:')
    for k, v in net.user_metadata['class_mapping']['baselines'].items():
        print(f'  {k}\t{v}')
    print('Training region types:')
    for k, v in net.user_metadata['class_mapping']['regions'].items():
        print(f'  {k}\t{v}')

    print(f'existing bounding regions: {net.user_metadata["bounding_regions"]}')

    if bounding_region:
        br = set(net.user_metadata["bounding_regions"])
        br_new = set(bounding_region)

        print(f'removing: {br.difference(br_new)}')
        print(f'adding: {br_new.difference(br)}')
        net.user_metadata["bounding_regions"] = bounding_region

    print(f'Model is {"topline" if "topline" in net.user_metadata and net.user_metadata["topline"] else "baseline"}')
    print(f'Setting to  {"topline" if topline else "baseline"}')
    net.user_metadata['topline'] = topline

    print(f"Model has padding {net.user_metadata['hyper_params']['padding'] if 'padding' in net.user_metadata['hyper_params'] else (0, 0)}")
    print(f'Setting to {pad}')
    net.user_metadata['hyper_params']['padding'] = pad

    shutil.copy(model, f'{model}.bak')
    net.save_model(model)


if __name__ == '__main__':
    cli()



#! /usr/bin/env python
"""
Produces semi-transparent neural segmenter output overlays
"""
import click


@click.command()
@click.argument('files', nargs=-1)
def cli(files):

    from os.path import splitext

    import torch
    import torchvision.transforms as tf
    from PIL import Image

    from kraken.lib import dataset

    batch, channels, height, width = 1, 3, 1200, 0
    transforms = dataset.ImageInputTransforms(batch, height, width, channels, 0, valid_norm=False)

    torch.set_num_threads(1)

    ds = dataset.BaselineSet(files, im_transforms=transforms, mode='xml')

    for idx, batch in enumerate(ds):
        img = ds.imgs[idx]
        print(img)
        im = Image.open(img)
        res_tf = tf.Compose(transforms.transforms[:2])
        scal_im = res_tf(im)
        o = batch['target'].numpy()
        heat = Image.fromarray((o[ds.class_mapping['baselines']['default']]*255).astype('uint8'))
        heat.save(splitext(img)[0] + '.heat.png')
        overlay = Image.new('RGBA', scal_im.size, (0, 130, 200, 255))
        bl = Image.composite(overlay, scal_im.convert('RGBA'), heat)
        heat = Image.fromarray((o[ds.class_mapping['aux']['_start_separator']]*255).astype('uint8'))
        overlay = Image.new('RGBA', scal_im.size, (230, 25, 75, 255))
        bl = Image.composite(overlay, bl, heat)
        heat = Image.fromarray((o[ds.class_mapping['aux']['_end_separator']]*255).astype('uint8'))
        overlay = Image.new('RGBA', scal_im.size, (60, 180, 75, 255))
        Image.composite(overlay, bl, heat).save(splitext(img)[0] + '.overlay.png')
        del o
        del im


if __name__ == '__main__':
    cli()



#
# Copyright 2015 Benjamin Kiessling
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
import datetime
import importlib.metadata
import logging
from typing import (TYPE_CHECKING, Any, Dict, Iterable, List, Literal,
                    Optional, Sequence, Tuple)

import regex
from jinja2 import Environment, FunctionLoader, PackageLoader

from kraken.lib.util import make_printable

if TYPE_CHECKING:
    from collections import Counter
    from os import PathLike

    from kraken.containers import ProcessingStep, Segmentation

logger = logging.getLogger(__name__)

__all__ = ['serialize', 'render_report']


def _rescale(val: Sequence[float], low: float, high: float) -> List[float]:
    """
    Rescales a list of confidence value between 0 and 1 to an interval [low,
    high].

    Args:
        val (float): List of values in interval (0,1)
        low (float): Lower bound of rescaling interval
        high (float): Upper bound of rescaling interval

    Returns:
        Rescaled value (float).
    """
    return [(high - low) * x + low for x in val]


def max_bbox(boxes: Iterable[Sequence[int]]) -> Tuple[int, int, int, int]:
    """
    Calculates the minimal bounding box containing all contained in an
    iterator.

    Args:
        boxes (iterator): An iterator returning tuples of the format ((x0, y0),
        (x1, y1), ... (xn, yn)).
    Returns:
        A box (x0, y0, x1, y1) covering all bounding boxes in the input
        argument.
    """
    flat_box = [point for pol in boxes for point in pol]
    flat_box = [x for point in flat_box for x in point]
    xmin, xmax = min(flat_box[::2]), max(flat_box[::2])
    ymin, ymax = min(flat_box[1::2]), max(flat_box[1::2])
    o = xmin, ymin, xmax, ymax  # type: ignore
    return o


def serialize(results: 'Segmentation',
              image_size: Tuple[int, int] = (0, 0),
              writing_mode: Literal['horizontal-tb', 'vertical-lr', 'vertical-rl'] = 'horizontal-tb',
              scripts: Optional[Iterable[str]] = None,
              template: ['PathLike', str] = 'alto',
              template_source: Literal['native', 'custom'] = 'native',
              processing_steps: Optional[List['ProcessingStep']] = None) -> str:
    """
    Serializes recognition and segmentation results into an output document.

    Serializes a Segmentation container object containing either segmentation
    or recognition results into an output document. The rendering is performed
    with jinja2 templates that can either be shipped with kraken
    (`template_source` == 'native') or custom (`template_source` == 'custom').

    Note: Empty records are ignored for serialization purposes.

    Args:
        segmentation: Segmentation container object
        image_size: Dimensions of the source image
        writing_mode: Sets the principal layout of lines and the
                      direction in which blocks progress. Valid values are
                      horizontal-tb, vertical-rl, and vertical-lr.
        scripts: List of scripts contained in the OCR records
        template: Selector for the serialization format. May be 'hocr',
                  'alto', 'page' or any template found in the template
                  directory. If template_source is set to `custom` a path to a
                  template is expected.
        template_source: Switch to enable loading of custom templates from
                         outside the kraken package.
        processing_steps: A list of ProcessingStep container classes describing
                          the processing kraken performed on the inputs.

    Returns:
        The rendered template
    """
    logger.info(f'Serialize {len(results.lines)} records from {results.imagename} with template {template}.')
    page: Dict[str, Any] = {'entities': [],
                            'size': image_size,
                            'name': results.imagename,
                            'writing_mode': writing_mode,
                            'scripts': scripts,
                            'date': datetime.datetime.now(datetime.timezone.utc).isoformat(),
                            'base_dir': [rec.base_dir for rec in results.lines][0] if len(results.lines) else None,
                            'seg_type': results.type}
    metadata = {'processing_steps': processing_steps,
                'version': importlib.metadata.version('kraken')}

    seg_idx = 0
    char_idx = 0

    # build region and line type dict
    types = []
    for line in results.lines:
        if line.tags is not None:
            types.extend((k, v) for k, v in line.tags.items())
    page['line_types'] = list(set(types))
    page['region_types'] = list(results.regions.keys())

    # map reading orders indices to line IDs
    ros = []
    for ro in results.line_orders:
        ros.append([results.lines[idx].id for idx in ro])
    page['line_orders'] = ros
    # build region ID to region dict
    reg_dict = {}
    for key, regs in results.regions.items():
        for reg in regs:
            reg_dict[reg.id] = reg

    regs_with_lines = set()
    prev_reg = None
    for idx, record in enumerate(results.lines):
        # line not in region
        if not record.regions or len(record.regions) == 0:
            cur_ent = page['entities']
        # line not in same region as previous line
        elif prev_reg != record.regions[0]:
            prev_reg = record.regions[0]
            reg = reg_dict[record.regions[0]]
            regs_with_lines.add(reg.id)
            region = {'id': reg.id,
                      'bbox': max_bbox([reg.boundary]) if reg.boundary else [],
                      'boundary': [list(x) for x in reg.boundary] if reg.boundary else [],
                      'tags': reg.tags,
                      'lines': [],
                      'type': 'region'
                      }
            page['entities'].append(region)
            cur_ent = region['lines']

        # set field to indicate the availability of baseline segmentation in
        # addition to bounding boxes
        line = {'id': record.id,
                'bbox': max_bbox([record.boundary]) if record.type == 'baselines' else record.bbox,
                'cuts': [list(x) for x in getattr(record, 'cuts', [])],
                'confidences': getattr(record, 'confidences', []),
                'recognition': [],
                'boundary': [list(x) for x in record.boundary] if record.type == 'baselines' else [[record.bbox[0], record.bbox[1]],
                                                                                                   [record.bbox[2], record.bbox[1]],
                                                                                                   [record.bbox[2], record.bbox[3]],
                                                                                                   [record.bbox[0], record.bbox[3]]],
                'type': 'line'
                }
        if record.tags is not None:
            line['tags'] = record.tags
        if record.type == 'baselines':
            line['baseline'] = [list(x) for x in record.baseline]

        splits = regex.split(r'(\s+)', getattr(record, 'prediction', ''))
        line_offset = 0
        logger.debug(f'Record contains {len(splits)} segments')
        for segment in splits:
            if len(segment) == 0:
                continue
            seg_cuts = record.cuts[line_offset:line_offset + len(segment)]
            seg_bbox = max_bbox(seg_cuts)
            seg_struct = {'bbox': seg_bbox,
                          'confidences': record.confidences[line_offset:line_offset + len(segment)],
                          'cuts': seg_cuts,
                          'text': segment,
                          'recognition': [{'bbox': max_bbox([cut]),
                                           'boundary': cut,
                                           'confidence': conf,
                                           'text': char,
                                           'index': cid}
                                          for conf, cut, char, cid in
                                          zip(record.confidences[line_offset:line_offset + len(segment)],
                                              seg_cuts,
                                              segment,
                                              range(char_idx, char_idx + len(segment)))],
                          'index': seg_idx}
            if record.type == 'baselines':
                seg_struct['boundary'] = record[line_offset:line_offset + len(segment)][1]
            line['recognition'].append(seg_struct)
            char_idx += len(segment)
            seg_idx += 1
            line_offset += len(segment)
        cur_ent.append(line)

    # serialize all remaining (line-less) regions
    for reg_id in regs_with_lines:
        reg_dict.pop(reg_id)
    logger.debug(f'No lines given but {len(results.regions)}. Serialize all regions.')
    for reg in reg_dict.values():
        region = {'id': reg.id,
                  'bbox': max_bbox([reg.boundary]),
                  'boundary': [list(x) for x in reg.boundary],
                  'tags': reg.tags,
                  'lines': [],
                  'type': 'region'
                  }
        page['entities'].append(region)

    if template_source == 'native':
        logger.debug('Initializing native jinja environment.')
        loader = PackageLoader('kraken', 'templates')
    elif template_source == 'custom':
        def _load_template(name):
            return open(template, 'r').read(), name, lambda: True
        loader = FunctionLoader(_load_template)

    env = Environment(loader=loader,
                      trim_blocks=True,
                      lstrip_blocks=True,
                      autoescape=True)
    env.tests['whitespace'] = str.isspace
    env.filters['rescale'] = _rescale
    logger.debug('Retrieving template.')
    tmpl = env.get_template(template)
    logger.debug('Rendering data.')
    return tmpl.render(page=page, metadata=metadata)


def render_report(model: str,
                  chars: int,
                  errors: int,
                  char_accuracy: float,
                  word_accuracy: float,
                  char_confusions: 'Counter',
                  scripts: 'Counter',
                  insertions: 'Counter',
                  deletions: int,
                  substitutions: 'Counter') -> str:
    """
    Renders an accuracy report.

    Args:
        model (str): Model name.
        errors (int): Number of errors on test set.
        char_confusions (dict): Dictionary mapping a tuple (gt, pred) to a
                                number of occurrences.
        scripts (dict): Dictionary counting character per script.
        insertions (dict): Dictionary counting insertion operations per Unicode
                           script
        deletions (int): Number of deletions
        substitutions (dict): Dictionary counting substitution operations per
                              Unicode script.

    Returns:
        A string containing the rendered report.
    """
    logger.info(f'Serializing report for {model}.')

    report = {'model': model,
              'chars': chars,
              'errors': errors,
              'character_accuracy': char_accuracy * 100,
              'word_accuracy': word_accuracy * 100,
              'insertions': sum(insertions.values()),
              'deletions': deletions,
              'substitutions': sum(substitutions.values()),
              'scripts': sorted([{'script': k,
                                  'count': v,
                                  'errors': insertions[k] + substitutions[k],
                                  'accuracy': 100 * (v-(insertions[k] + substitutions[k]))/v} for k, v in scripts.items()],
                                key=lambda x: x['accuracy'],
                                reverse=True),
              'counts': sorted([{'correct': make_printable(k[0]),
                                 'generated': make_printable(k[1]),
                                 'errors': v} for k, v in char_confusions.items() if k[0] != k[1]],
                               key=lambda x: x['errors'],
                               reverse=True)}
    logger.debug('Initializing jinja environment.')
    env = Environment(loader=PackageLoader('kraken', 'templates'),
                      trim_blocks=True,
                      lstrip_blocks=True,
                      autoescape=True)
    logger.debug('Retrieving template.')
    tmpl = env.get_template('report')
    logger.debug('Rendering data.')
    return tmpl.render(report=report)



#
# Copyright 2015 Benjamin Kiessling
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
"""
Arrow IPC format dataset builders.
"""
__all__ = ['build_binary_dataset']

import io
import json
import tempfile
from collections import Counter
from functools import partial
from multiprocessing import Pool
from typing import TYPE_CHECKING, Literal, Callable, List, Optional, Tuple, Union

import numpy as np
import pyarrow as pa
from PIL import Image, UnidentifiedImageError

from kraken.containers import Segmentation
from kraken.lib import functional_im_transforms as F_t
from kraken.lib.exceptions import KrakenInputException
from kraken.lib.segmentation import extract_polygons
from kraken.lib.util import is_bitonal, make_printable
from kraken.lib.xml import XMLPage

if TYPE_CHECKING:
    from os import PathLike

import logging

logger = logging.getLogger(__name__)


def _extract_line(xml_record, skip_empty_lines: bool = True, legacy_polygons: bool = False):
    lines = []
    try:
        im = Image.open(xml_record.imagename)
    except (FileNotFoundError, UnidentifiedImageError):
        return lines, None, None
    if is_bitonal(im):
        im = im.convert('1')
    for idx, rec in enumerate(xml_record.lines):
        seg = Segmentation(text_direction='horizontal-lr',
                           imagename=xml_record.imagename,
                           type=xml_record.type,
                           lines=[rec],
                           regions=None,
                           script_detection=False,
                           line_orders=[])
        try:
            line_im, line = next(extract_polygons(im, seg, legacy=legacy_polygons))
        except KrakenInputException:
            logger.warning(f'Invalid line {idx} in {xml_record.imagename}')
            continue
        except Exception as e:
            logger.warning(f'Unexpected exception {e} from line {idx} in {xml_record.imagename}')
            continue
        if not line.text and skip_empty_lines:
            continue
        fp = io.BytesIO()
        line_im.save(fp, format='png')
        lines.append({'text': line.text, 'im': fp.getvalue()})
    return lines, im.mode


def _extract_path_line(xml_record, skip_empty_lines: bool = True):
    try:
        im = Image.open(xml_record['image'])
    except (FileNotFoundError, UnidentifiedImageError):
        return [], None, None
    if not xml_record['lines'][0]['text'] and skip_empty_lines:
        return [], None, None
    if is_bitonal(im):
        im = im.convert('1')
    fp = io.BytesIO()
    im.save(fp, format='png')
    line = {'text': xml_record['lines'][0]['text'], 'im': fp.getvalue()}
    return [line], im.mode


def parse_path(path: Union[str, 'PathLike'],
               suffix: str = '.gt.txt',
               split=F_t.default_split,
               skip_empty_lines: bool = True):
    with open(F_t.suffix_split(path, split=split, suffix=suffix), 'r', encoding='utf-8') as fp:
        gt = fp.read().strip('\n\r')
        if not gt and skip_empty_lines:
            raise KrakenInputException(f'No text for ground truth line {path}.')

    return {'image': path, 'lines': [{'text': gt}]}


def build_binary_dataset(files: Optional[List[Union[str, 'PathLike', 'Segmentation']]] = None,
                         output_file: Union[str, 'PathLike'] = None,
                         format_type: Literal['xml', 'alto', 'page', None] = 'xml',
                         num_workers: int = 0,
                         ignore_splits: bool = False,
                         random_split: Optional[Tuple[float, float, float]] = None,
                         force_type: Optional[str] = None,
                         recordbatch_size: int = 100,
                         skip_empty_lines: bool = True,
                         callback: Callable[[int, int], None] = lambda chunk, lines: None,
                         legacy_polygons: bool = False) -> None:
    """
    Parses XML files and dumps the baseline-style line images and text into a
    binary dataset.

    Args:
        files: List of XML input files or Segmentation container objects.
        output_file: Path to the output file.
        format_type: One of `xml`, `alto`, `page`, `path`, or None. In `None`
                     mode, the files argument is expected to be a list of
                     `kraken.containers.Segmentation` objects.
        num_workers: Number of workers for parallelized extraction of line
                     images. Set to `0` to disable parallelism.
        ignore_splits: Switch to disable serialization of the explicit
                       train/validation/test splits contained in the source
                       files.
        random_split: Serializes a random split into the dataset with the
                       proportions (train, val, test).
        force_type: Forces a dataset type. Can be `kraken_recognition_baseline`
                    or `kraken_recognition_bbox`.
        recordbatch_size: Minimum number of records per RecordBatch written to
                          the output file. Larger batches require more
                          transient memory but slightly improve reading
                          performance.
        skip_empty_lines: Do not compile empty text lines into the dataset.
        callback: Function called every time a new recordbatch is flushed into
                  the Arrow IPC file.
        legacy_polygons: Use legacy polygon extraction code.
    """

    logger.info('Parsing XML files')
    extract_fn = partial(_extract_line, skip_empty_lines=skip_empty_lines, legacy_polygons=legacy_polygons)
    parse_fn = None
    if format_type in ['xml', 'alto', 'page']:
        parse_fn = XMLPage
    elif format_type == 'path':
        if not ignore_splits:
            logger.warning('ignore_splits is False and format_type is path. Will not serialize splits.')
        parse_fn = partial(parse_path, skip_empty_lines=skip_empty_lines)
        extract_fn = partial(_extract_path_line, skip_empty_lines=skip_empty_lines)
    elif format_type is None:
        pass
    else:
        raise ValueError(f'invalid format {format_type} for parse_(xml,alto,page,path)')

    if force_type and force_type not in ['kraken_recognition_baseline', 'kraken_recognition_bbox']:
        raise ValueError(f'force_type set to invalid value {force_type}')

    docs = []
    if parse_fn:
        for doc in files:
            try:
                data = parse_fn(doc)
                if format_type in ['xml', 'alto', 'page']:
                    data = data.to_container()
            except (FileNotFoundError, KrakenInputException, ValueError):
                logger.warning(f'Invalid input file {doc}')
                continue
            try:
                if format_type in ['xml', 'alto', 'page']:
                    imagename = data.imagename
                else:
                    imagename = data['image']
                with open(imagename, 'rb') as fp:
                    Image.open(fp)
            except (FileNotFoundError, UnidentifiedImageError) as e:
                logger.warning(f'Could not open file {e.filename} in {doc}')
                continue
            docs.append(data)
        logger.info(f'Parsed {len(docs)} files.')
    else:
        docs = files.copy()
        logger.info(f'Got {len(docs)} preparsed files.')

    logger.info('Assembling dataset alphabet.')
    alphabet = Counter()
    num_lines = 0
    for doc in docs:
        if format_type in ['xml', 'alto', 'page', None]:
            lines = doc.lines
        elif format_type == 'path':
            lines = doc['lines']
        for line in lines:
            num_lines += 1
            alphabet.update(line.text if format_type in ['xml', 'alto', 'page', None] else line['text'])

    callback(0, num_lines)

    for k, v in sorted(alphabet.items(), key=lambda x: x[1], reverse=True):
        char = make_printable(k)
        if char == k:
            char = '\t' + char
        logger.info(f'{char}\t{v}')

    if force_type:
        ds_type = force_type
    else:
        ds_type = 'kraken_recognition_baseline' if format_type != 'path' else 'kraken_recognition_bbox'

    metadata = {'lines': {'type': ds_type,
                          'alphabet': alphabet,
                          'text_type': 'raw',
                          'image_type': 'raw',
                          'splits': ['train', 'eval', 'test'],
                          'im_mode': '1',
                          'legacy_polygons': legacy_polygons,
                          'counts': Counter({'all': 0,
                                             'train': 0,
                                             'validation': 0,
                                             'test': 0
                                             }
                                            ),
                          }
                }

    ty = pa.struct([('text', pa.string()), ('im', pa.binary())])
    schema = pa.schema([('lines', ty), ('train', pa.bool_()), ('validation', pa.bool_()), ('test', pa.bool_())])

    def _make_record_batch(line_cache):
        ar = pa.array(line_cache, type=ty)
        if random_split:
            indices = np.random.choice(4, len(line_cache), p=(0.0,) + random_split)
        else:
            indices = np.zeros(len(line_cache))
        tr_ind = np.zeros(len(line_cache), dtype=bool)
        tr_ind[indices == 1] = True
        val_ind = np.zeros(len(line_cache), dtype=bool)
        val_ind[indices == 2] = True
        test_ind = np.zeros(len(line_cache), dtype=bool)
        test_ind[indices == 3] = True

        train_mask = pa.array(tr_ind)
        val_mask = pa.array(val_ind)
        test_mask = pa.array(test_ind)
        rbatch = pa.RecordBatch.from_arrays([ar, train_mask, val_mask, test_mask], schema=schema)
        return rbatch, (len(line_cache), int(sum(indices == 1)), int(sum(indices == 2)), int(sum(indices == 3)))

    line_cache = []
    logger.info('Writing lines to temporary file.')
    with tempfile.TemporaryDirectory() as tmp_output_dir:
        tmp_file = tmp_output_dir + '/dataset.arrow'
        with pa.OSFile(tmp_file, 'wb') as sink:
            with pa.ipc.new_file(sink, schema) as writer:

                if num_workers and num_workers > 1:
                    logger.info(f'Spinning up processing pool with {num_workers} workers.')
                    with Pool(num_workers) as pool:
                        for page_lines, im_mode in pool.imap_unordered(extract_fn, docs):
                            if page_lines:
                                line_cache.extend(page_lines)
                                # comparison RGB(A) > L > 1
                                if im_mode > metadata['lines']['im_mode']:
                                    metadata['lines']['im_mode'] = im_mode

                            if len(line_cache) >= recordbatch_size:
                                logger.info(f'Flushing {len(line_cache)} lines into {tmp_file}.')
                                rbatch, counts = _make_record_batch(line_cache)
                                metadata['lines']['counts'].update({'all': counts[0],
                                                                    'train': counts[1],
                                                                    'validation': counts[2],
                                                                    'test': counts[3]})
                                writer.write(rbatch)
                                callback(len(line_cache), num_lines)
                                line_cache = []
                else:
                    for page_lines, im_mode in map(extract_fn, docs):
                        if page_lines:
                            line_cache.extend(page_lines)
                            # comparison RGB(A) > L > 1
                            if im_mode > metadata['lines']['im_mode']:
                                metadata['lines']['im_mode'] = im_mode

                        if len(line_cache) >= recordbatch_size:
                            logger.info(f'Flushing {len(line_cache)} lines into {tmp_file}.')
                            rbatch, counts = _make_record_batch(line_cache)
                            metadata['lines']['counts'].update({'all': counts[0],
                                                                'train': counts[1],
                                                                'validation': counts[2],
                                                                'test': counts[3]})
                            writer.write(rbatch)
                            callback(len(line_cache), num_lines)
                            line_cache = []

                if line_cache:
                    logger.info(f'Flushing last {len(line_cache)} lines into {tmp_file}.')
                    rbatch, counts = _make_record_batch(line_cache)
                    metadata['lines']['counts'].update({'all': counts[0],
                                                        'train': counts[1],
                                                        'validation': counts[2],
                                                        'test': counts[3]})
                    writer.write(rbatch)
                    callback(len(line_cache), num_lines)

        logger.info('Dataset metadata')
        logger.info(f"type: {metadata['lines']['type']}\n"
                    f"text_type: {metadata['lines']['text_type']}\n"
                    f"image_type: {metadata['lines']['image_type']}\n"
                    f"splits: {metadata['lines']['splits']}\n"
                    f"im_mode: {metadata['lines']['im_mode']}\n"
                    f"legacy_polygons: {metadata['lines']['legacy_polygons']}\n"
                    f"lines: {metadata['lines']['counts']}\n")

        with pa.memory_map(tmp_file, 'rb') as source:
            logger.info(f'Rewriting output ({output_file}) to update metadata.')
            ds = pa.ipc.open_file(source).read_all()
            metadata['lines']['counts'] = dict(metadata['lines']['counts'])
            metadata['lines'] = json.dumps(metadata['lines'])
            schema = schema.with_metadata(metadata)
            with pa.OSFile(output_file, 'wb') as sink:
                with pa.ipc.new_file(sink, schema) as writer:
                    writer.write(ds)



#
# Copyright 2015 Benjamin Kiessling
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
"""
kraken.rpred
~~~~~~~~~~~~

Generators for recognition on lines images.
"""
import dataclasses
import logging
from collections import defaultdict
from functools import partial
from typing import (TYPE_CHECKING, Dict, Generator, List, Optional, Sequence,
                    Tuple, Union)
import warnings

from kraken.containers import BaselineOCRRecord, BBoxOCRRecord, ocr_record
from kraken.lib.dataset import ImageInputTransforms
from kraken.lib.exceptions import KrakenInputException
from kraken.lib.segmentation import extract_polygons
from kraken.lib.util import get_im_str, is_bitonal

if TYPE_CHECKING:
    from PIL import Image

    from kraken.containers import Segmentation
    from kraken.lib.models import TorchSeqRecognizer

__all__ = ['mm_rpred', 'rpred']

logger = logging.getLogger(__name__)


class mm_rpred(object):
    """
    Multi-model version of kraken.rpred.rpred
    """
    def __init__(self,
                 nets: Dict[Tuple[str, str], 'TorchSeqRecognizer'],
                 im: 'Image.Image',
                 bounds: 'Segmentation',
                 pad: int = 16,
                 bidi_reordering: Union[bool, str] = True,
                 tags_ignore: Optional[List[Tuple[str, str]]] = None,
                 no_legacy_polygons: bool = False) -> Generator[ocr_record, None, None]:
        """
        Multi-model version of kraken.rpred.rpred.

        Takes a dictionary of ISO15924 script identifiers->models and an
        script-annotated segmentation to dynamically select appropriate models for
        these lines.

        Args:
            nets: A dict mapping tag key-value pairs to TorchSegRecognizer
                  objects. Recommended to be an defaultdict.
            im: Image to extract text from
            bounds: A Segmentation data class containing either bounding box or
                    baseline type segmentation.
            pad: Extra blank padding to the left and right of text line
            bidi_reordering: Reorder classes in the ocr_record according to the
                             Unicode bidirectional algorithm for correct
                             display. Set to L|R to override default text
                             direction.
            tags_ignore: List of tag key-value pairs to ignore during
                         recognition

        Yields:
            An ocr_record containing the recognized text, absolute character
            positions, and confidence values for each character.

        Raises:
            KrakenInputException if the mapping between segmentation tags and
            networks is incomplete.
        """
        seg_types = set(recognizer.seg_type for recognizer in nets.values())
        if isinstance(nets, defaultdict):
            seg_types.add(nets.default_factory().seg_type)
            self._resolve_tags_to_model = partial(_resolve_tags_to_model, default=nets.default_factory())
        else:
            self._resolve_tags_to_model = _resolve_tags_to_model

        if not tags_ignore:
            tags_ignore = []

        if bounds.script_detection:
            self.have_tags = True
        else:
            self.have_tags = False

        if bounds.type not in seg_types or len(seg_types) > 1:
            logger.warning(f'Recognizers with segmentation types {seg_types} will be '
                           f'applied to segmentation of type {bounds.type}. '
                           f'This will likely result in severely degraded performace')
        one_channel_modes = set(recognizer.nn.one_channel_mode for recognizer in nets.values())
        if '1' in one_channel_modes and len(one_channel_modes) > 1:
            raise ValueError('Mixing binary and non-binary recognition models is not supported.')
        elif '1' in one_channel_modes and not is_bitonal(im):
            logger.warning('Running binary models on non-binary input image '
                           f'(mode {im.mode}). This will result in severely degraded '
                           'performance')

        self.len = len(bounds.lines)
        self.line_iter = iter(bounds.lines)

        if bounds.type == 'baselines':
            valid_norm = False
            self.next_iter = self._recognize_baseline_line
        else:
            valid_norm = True
            self.next_iter = self._recognize_box_line

        if isinstance(nets, defaultdict) and nets.default_factory:
            network = nets.default_factory()
            batch, channels, height, width = network.nn.input
            self.ts = defaultdict(lambda: ImageInputTransforms(batch, height, width, channels, (pad, 0), valid_norm))
        else:
            self.ts = {}

        if self.have_tags:
            tags = set()
            for x in bounds.lines:
                tags.update(x.tags.items())

            im_str = get_im_str(im)
            logger.info(f'Running {len(nets)} multi-script recognizers on {im_str} with {self.len} lines')

            filtered_tags = []
            miss = []
            for tag in tags:
                if not isinstance(nets, defaultdict) and (not nets.get(tag) and tag not in tags_ignore):
                    miss.append(tag)
                elif tag not in tags_ignore:
                    filtered_tags.append(tag)
            tags = filtered_tags

            if miss:
                raise KrakenInputException(f'Missing models for tags {set(miss)}')

            # build dictionary for line preprocessing
            self.ts = {}
            for tag in tags:
                logger.debug(f'Loading line transforms for {tag}')
                network = nets[tag]
                batch, channels, height, width = network.nn.input
                self.ts[tag] = ImageInputTransforms(batch, height, width, channels, (pad, 0), valid_norm)

        if not isinstance(self.ts, defaultdict) and not self.ts:
            raise ValueError('No tags in input data and no default model in mapping given.')

        self.im = im
        self.nets = nets
        self.bidi_reordering = bidi_reordering
        self.pad = pad
        self.bounds = bounds
        self.tags_ignore = tags_ignore
        self.no_legacy_polygons = no_legacy_polygons

    def _recognize_box_line(self, line):
        xmin, ymin, xmax, ymax = line.bbox
        prediction = ''
        cuts = []
        confidences = []
        line.text_direction = self.bounds.text_direction

        if self.have_tags and self.tags_ignore:
            for tag in line.tags.items():
                if tag in self.tags_ignore:
                    logger.info(f'Ignoring line segment with tags {line.tags} based on {tag}.')
                    return BBoxOCRRecord('', (), (), line)

        tag, net = self._resolve_tags_to_model(line.tags, self.nets)

        use_legacy_polygons = self._choose_legacy_polygon_extractor(net)

        seg = dataclasses.replace(self.bounds, lines=[line])
        box, coords = next(extract_polygons(self.im, seg, legacy=use_legacy_polygons))
        self.box = box

        # check if boxes are non-zero in any dimension
        if 0 in box.size:
            logger.warning(f'bbox {line} with zero dimension. Emitting empty record.')
            return BBoxOCRRecord('', (), (), line)
        # try conversion into tensor
        try:
            logger.debug('Preparing run.')
            ts_box = self.ts[tag](box)
        except Exception:
            logger.warning(f'Conversion of line {line} failed. Emitting empty record..')
            return BBoxOCRRecord('', (), (), line)

        # check if line is non-zero
        if ts_box.max() == ts_box.min():
            logger.warning('Empty run. Emitting empty record.')
            return BBoxOCRRecord('', (), (), line)

        logger.debug(f'Forward pass with model {tag}.')
        preds = net.predict(ts_box.unsqueeze(0))[0]

        # calculate recognized LSTM locations of characters
        logger.debug('Convert to absolute coordinates')
        # calculate recognized LSTM locations of characters
        # scale between network output and network input
        self.net_scale = ts_box.shape[2]/net.outputs.shape[2]
        # scale between network input and original line
        self.in_scale = box.size[0]/(ts_box.shape[2]-2*self.pad)

        pred = ''.join(x[0] for x in preds)
        pos = []
        conf = []

        for _, start, end, c in preds:
            if self.bounds.text_direction.startswith('horizontal'):
                xmin = coords[0] + self._scale_val(start, 0, self.box.size[0])
                xmax = coords[0] + self._scale_val(end, 0, self.box.size[0])
                pos.append([[xmin, coords[1]], [xmin, coords[3]], [xmax, coords[3]], [xmax, coords[1]]])
            else:
                ymin = coords[1] + self._scale_val(start, 0, self.box.size[1])
                ymax = coords[1] + self._scale_val(end, 0, self.box.size[1])
                pos.append([[coords[0], ymin], [coords[2], ymin], [coords[2], ymax], [coords[0], ymax]])
            conf.append(c)
        prediction += pred
        cuts.extend(pos)
        confidences.extend(conf)

        rec = BBoxOCRRecord(prediction, cuts, confidences, line)
        if self.bidi_reordering:
            logger.debug('BiDi reordering record.')
            return rec.logical_order(base_dir=self.bidi_reordering if self.bidi_reordering in ('L', 'R') else None)
        else:
            logger.debug('Emitting raw record')
            return rec.display_order(None)

    def _recognize_baseline_line(self, line):
        if self.have_tags and self.tags_ignore is not None:
            for tag in line.tags.items():
                if tag in self.tags_ignore:
                    logger.info(f'Ignoring line segment with tags {line.tags} based on {tag}.')
                    return BaselineOCRRecord('', [], [], line)

        seg = dataclasses.replace(self.bounds, lines=[line])

        tag, net = self._resolve_tags_to_model(line.tags, self.nets)

        use_legacy_polygons = self._choose_legacy_polygon_extractor(net)

        try:
            box, coords = next(extract_polygons(self.im, seg, legacy=use_legacy_polygons))
        except KrakenInputException as e:
            logger.warning(f'Extracting line failed: {e}')
            return BaselineOCRRecord('', [], [], line)

        self.box = box
        # check if boxes are non-zero in any dimension
        if 0 in box.size:
            logger.warning(f'{line} with zero dimension. Emitting empty record.')
            return BaselineOCRRecord('', [], [], line)
        # try conversion into tensor
        try:
            ts_box = self.ts[tag](box)
        except Exception as e:
            logger.warning(f'Tensor conversion failed with {e}. Emitting empty record.')
            return BaselineOCRRecord('', [], [], line)
        # check if line is non-zero
        if ts_box.max() == ts_box.min():
            logger.warning('Empty line after tensor conversion. Emitting empty record.')
            return BaselineOCRRecord('', [], [], line)

        preds = net.predict(ts_box.unsqueeze(0))[0]
        # calculate recognized LSTM locations of characters
        # scale between network output and network input
        self.net_scale = ts_box.shape[2]/net.outputs.shape[2]
        # scale between network input and original line
        self.in_scale = box.size[0]/(ts_box.shape[2]-2*self.pad)

        # XXX: fix bounding box calculation ocr_record for multi-codepoint labels.
        pred = ''.join(x[0] for x in preds)
        pos = []
        conf = []
        for _, start, end, c in preds:
            pos.append([self._scale_val(start, 0, self.box.size[0]),
                        self._scale_val(end, 0, self.box.size[0])])
            conf.append(c)
        rec = BaselineOCRRecord(pred, pos, conf, line)
        if self.bidi_reordering:
            logger.debug('BiDi reordering record.')
            return rec.logical_order(base_dir=self.bidi_reordering if self.bidi_reordering in ('L', 'R') else None)
        else:
            logger.debug('Emitting raw record')
            return rec.display_order(None)

    def __next__(self):
        return self.next_iter(next(self.line_iter))

    def __iter__(self):
        return self

    def __len__(self):
        return self.len

    def _scale_val(self, val, min_val, max_val):
        return int(round(min(max(((val*self.net_scale)-self.pad)*self.in_scale, min_val), max_val-1)))

    def _choose_legacy_polygon_extractor(self, net) -> bool:
        # grouping the checks here to display warnings only once
        if net.nn.use_legacy_polygons:
            if self.no_legacy_polygons:
                warnings.warn('Enforcing use of the new polygon extractor for models trained with old version. Accuracy may be affected.')
                return False
            else:
                warnings.warn('Using legacy polygon extractor, as the model was not trained with the new method. Please retrain your model to get speed improvement.')
                return True
        return False



def rpred(network: 'TorchSeqRecognizer',
          im: 'Image.Image',
          bounds: 'Segmentation',
          pad: int = 16,
          bidi_reordering: Union[bool, str] = True,
          no_legacy_polygons: bool = False) -> Generator[ocr_record, None, None]:
    """
    Uses a TorchSeqRecognizer and a segmentation to recognize text

    Args:
        network: A TorchSegRecognizer object
        im: Image to extract text from
        bounds: A Segmentation class instance containing either a baseline or
                bbox segmentation.
        pad: Extra blank padding to the left and right of text line.
             Auto-disabled when expected network inputs are incompatible with
             padding.
        bidi_reordering: Reorder classes in the ocr_record according to the
                         Unicode bidirectional algorithm for correct display.
                         Set to L|R to change base text direction.

    Yields:
        An ocr_record containing the recognized text, absolute character
        positions, and confidence values for each character.
    """
    return mm_rpred(defaultdict(lambda: network), im, bounds, pad, bidi_reordering, no_legacy_polygons=no_legacy_polygons)


def _resolve_tags_to_model(tags: Optional[Sequence[Dict[str, str]]],
                           model_map: Dict[Tuple[str, str], 'TorchSeqRecognizer'],
                           default: Optional['TorchSeqRecognizer'] = None) -> 'TorchSeqRecognizer':
    """
    Resolves a sequence of tags
    """
    if not tags and default:
        return ('type', 'default'), default
    elif tags:
        for tag in tags.items():
            if tag in model_map:
                return tag, model_map[tag]
    if tags and default:
        return next(iter(tags.items())), default
    raise KrakenInputException(f'No model for tags {tags}')



# -*- coding: utf-8 -*-
import unittest
from pathlib import Path

from PIL import Image
from pytest import raises

from kraken.lib import lineest

thisfile = Path(__file__).resolve().parent
resources = thisfile / 'resources'

class TestLineest(unittest.TestCase):

    """
    Testing centerline estimator
    """

    def setUp(self):
        self.lnorm = lineest.CenterNormalizer()

    def test_dewarp_bw(self):
        """
        Test dewarping of a single line in B/W
        """
        with Image.open(resources / '000236.png') as im:
            o = lineest.dewarp(self.lnorm, im.convert('1'))
            self.assertEqual(self.lnorm.target_height, o.size[1])

    def test_dewarp_gray(self):
        """
        Test dewarping of a single line in grayscale
        """
        with Image.open(resources /'000236.png') as im:
            o = lineest.dewarp(self.lnorm, im.convert('L'))
            self.assertEqual(self.lnorm.target_height, o.size[1])

    def test_dewarp_fail_color(self):
        """
        Test dewarping of a color line fails
        """
        with raises(ValueError):
            with Image.open(resources /'000236.png') as im:
                lineest.dewarp(self.lnorm, im.convert('RGB'))

    def test_dewarp_bw_undewarpable(self):
        """
        Test dewarping of an undewarpable line.
        """
        with Image.open(resources /'ONB_ibn_19110701_010.tif_line_1548924556947_449.png') as im:
            o = lineest.dewarp(self.lnorm, im)
            self.assertEqual(self.lnorm.target_height, o.size[1])




#
# Copyright 2019 Benjamin Kiessling
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
"""
ALTO/Page data loaders for segmentation training
"""
import logging
import re
from collections import defaultdict
from itertools import groupby
from pathlib import Path
from typing import (TYPE_CHECKING, Any, Dict, List, Literal, Optional,
                    Sequence, Tuple, Union)

from lxml import etree

from kraken.containers import BaselineLine, Region, Segmentation

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from os import PathLike

__all__ = ['XMLPage']


# fallback mapping between PAGE region types and tags
page_regions = {'TextRegion': 'text',
                'ImageRegion': 'image',
                'LineDrawingRegion': 'line drawing',
                'GraphicRegion': 'graphic',
                'TableRegion': 'table',
                'ChartRegion': 'chart',
                'MapRegion': 'map',
                'SeparatorRegion': 'separator',
                'MathsRegion': 'maths',
                'ChemRegion': 'chem',
                'MusicRegion': 'music',
                'AdvertRegion': 'advert',
                'NoiseRegion': 'noise',
                'UnknownRegion': 'unknown',
                'CustomRegion': 'custom'}

# same for ALTO
alto_regions = {'TextBlock': 'text',
                'Illustration': 'illustration',
                'GraphicalElement': 'graphic',
                'ComposedBlock': 'composed'}


class XMLPage(object):

    type: Literal['baselines', 'bbox'] = 'baselines'
    base_dir: Optional[Literal['L', 'R']] = None
    imagename: 'PathLike' = None
    image_size: Tuple[int, int] = None
    _orders: Dict[str, Dict[str, Any]] = None
    has_tags: bool = False
    _tag_set: Optional[Dict] = None
    has_splits: bool = False
    _split_set: Optional[List] = None

    def __init__(self,
                 filename: Union[str, 'PathLike'],
                 filetype: Literal['xml', 'alto', 'page'] = 'xml'):
        super().__init__()
        self.filename = Path(filename)
        self.filetype = filetype

        self._regions = {}
        self._lines = {}
        self._orders = {'line_implicit': {'order': [], 'is_total': True, 'description': 'Implicit line order derived from element sequence'},
                        'region_implicit': {'order': [], 'is_total': True, 'description': 'Implicit region order derived from element sequence'}}

        if filetype == 'xml':
            self._parse_xml()
        elif filetype == 'alto':
            self._parse_alto()
        elif filetype == 'page':
            self._parse_page()

    def _parse_xml(self):
        with open(self.filename, 'rb') as fp:
            try:
                doc = etree.parse(fp)
            except etree.XMLSyntaxError as e:
                raise ValueError(f'Parsing {self.filename} failed: {e}')
        if doc.getroot().tag.endswith('alto'):
            return self._parse_alto()
        elif doc.getroot().tag.endswith('PcGts'):
            return self._parse_page()
        else:
            raise ValueError(f'Unknown XML format in {self.filename}')

    def _parse_alto(self):
        with open(self.filename, 'rb') as fp:
            base_directory = self.filename.parent
            try:
                doc = etree.parse(fp)
            except etree.XMLSyntaxError as e:
                raise ValueError('Parsing {} failed: {}'.format(self.filename, e))
            image = doc.find('.//{*}fileName')
            if image is None or not image.text:
                raise ValueError('No valid image filename found in ALTO file {self.filename}')

            self.imagename = base_directory.joinpath(image.text)
            page = doc.find('.//{*}Page')
            self.image_size = int(page.get('WIDTH')), int(page.get('HEIGHT'))

            # find all image regions in order
            regions = []
            for el in doc.iterfind('./{*}Layout/{*}Page/{*}PrintSpace/{*}*'):
                for block_type in alto_regions.keys():
                    if el.tag.endswith(block_type):
                        regions.append(el)
            # find overall dimensions to filter out dummy TextBlocks
            ps = doc.find('./{*}Layout/{*}Page/{*}PrintSpace')
            x_min = int(float(ps.get('HPOS')))
            y_min = int(float(ps.get('VPOS')))
            width = int(float(ps.get('WIDTH')))
            height = int(float(ps.get('HEIGHT')))

            # parse tagrefs
            cls_map = {}
            tags = doc.find('.//{*}Tags')
            if tags is not None:
                for x in ['StructureTag', 'LayoutTag', 'OtherTag']:
                    for tag in tags.findall('./{{*}}{}'.format(x)):
                        cls_map[tag.get('ID')] = (x[:-3].lower(), tag.get('LABEL'))

            self._tag_set = set(('default',))

            # parse region type and coords
            region_data = defaultdict(list)
            for region in regions:
                region_id = region.get('ID')
                # try to find shape object
                coords = region.find('./{*}Shape/{*}Polygon')
                boundary = None
                if coords is not None:
                    boundary = self._parse_alto_pointstype(coords.get('POINTS'))
                elif (region.get('HPOS') is not None and region.get('VPOS') is not None and
                      region.get('WIDTH') is not None and region.get('HEIGHT') is not None):
                    # use rectangular definition
                    x_min = int(float(region.get('HPOS')))
                    y_min = int(float(region.get('VPOS')))
                    width = int(float(region.get('WIDTH')))
                    height = int(float(region.get('HEIGHT')))
                    boundary = [(x_min, y_min),
                                (x_min, y_min + height),
                                (x_min + width, y_min + height),
                                (x_min + width, y_min)]
                rtype = region.get('TYPE')
                # fall back to default region type if nothing is given
                tagrefs = region.get('TAGREFS')
                if tagrefs is not None and rtype is None:
                    for tagref in tagrefs.split():
                        ttype, rtype = cls_map.get(tagref, (None, None))
                        if rtype is not None and ttype:
                            break
                if rtype is None:
                    rtype = alto_regions[region.tag.split('}')[-1]]
                region_data[rtype].append(Region(id=region_id, boundary=boundary, tags={'type': rtype}))
                # register implicit reading order
                self._orders['region_implicit']['order'].append(region_id)

                # parse lines in region
                for line in region.iterfind('./{*}TextLine'):
                    if line.get('BASELINE') is None:
                        logger.info('TextLine {} without baseline'.format(line.get('ID')))
                        continue
                    pol = line.find('./{*}Shape/{*}Polygon')
                    boundary = None
                    if pol is not None:
                        try:
                            boundary = self._parse_alto_pointstype(pol.get('POINTS'))
                        except ValueError:
                            logger.info('TextLine {} without polygon'.format(line.get('ID')))
                    else:
                        logger.info('TextLine {} without polygon'.format(line.get('ID')))

                    baseline = None
                    try:
                        baseline = self._parse_alto_pointstype(line.get('BASELINE'))
                    except ValueError:
                        logger.info('TextLine {} without baseline'.format(line.get('ID')))

                    text = ''
                    for el in line.xpath(".//*[local-name() = 'String'] | .//*[local-name() = 'SP']"):
                        text += el.get('CONTENT') if el.get('CONTENT') else ' '
                    # find line type
                    tags = {'type': 'default'}
                    split_type = None
                    tagrefs = line.get('TAGREFS')
                    if tagrefs is not None:
                        for tagref in tagrefs.split():
                            ttype, ltype = cls_map.get(tagref, (None, None))
                            if ltype is not None:
                                self._tag_set.add(ltype)
                                if ttype == 'other':
                                    tags['type'] = ltype
                                else:
                                    tags[ttype] = ltype
                            if ltype in ['train', 'validation', 'test']:
                                split_type = ltype
                    self._lines[line.get('ID')] = BaselineLine(id=line.get('ID'),
                                                               baseline=baseline,
                                                               boundary=boundary,
                                                               text=text,
                                                               tags=tags,
                                                               split=split_type,
                                                               regions=[region_id])
                    # register implicit reading order
                    self._orders['line_implicit']['order'].append(line.get('ID'))

            self._regions = region_data

            if len(self._tag_set) > 1:
                self.has_tags = True
            else:
                self.has_tags = False

            # parse explicit reading orders if they exist
            ro_el = doc.find('.//{*}ReadingOrder')
            if ro_el is not None:
                reading_orders = ro_el.getchildren()
                # UnorderedGroup at top-level => treated as multiple reading orders
                if len(reading_orders) == 1 and reading_orders[0].tag.endswith('UnorderedGroup'):
                    reading_orders = reading_orders[0].getchildren()
                else:
                    reading_orders = [reading_orders]

                def _parse_group(el):
                    nonlocal is_valid

                    _ro = []
                    if el.tag.endswith('UnorderedGroup'):
                        _ro = [_parse_group(x) for x in el.iterchildren()]
                        is_total = False  # NOQA
                    elif el.tag.endswith('OrderedGroup'):
                        _ro.extend(_parse_group(x) for x in el.iterchildren())
                    else:
                        ref = el.get('REF')
                        res = doc.find(f'.//{{*}}*[@ID="{ref}"]')
                        if res is None:
                            logger.warning(f'Nonexistent element with ID {ref} in reading order. Skipping RO {ro.get("ID")}.')
                            is_valid = False
                            return _ro
                        tag = res.tag.split('}')[-1]
                        if tag not in alto_regions.keys() and tag != 'TextLine':
                            logger.warning(f'Sub-line element with ID {ref} in reading order. Skipping RO {ro.get("ID")}.')
                            is_valid = False
                            return _ro
                        return ref
                    return _ro

                for ro in reading_orders:
                    is_total = True
                    is_valid = True
                    joint_order = _parse_group(ro)
                    if is_valid:
                        tag = ro.get('TAGREFS')
                        self._orders[ro.get('ID')] = {'order': joint_order,
                                                      'is_total': is_total,
                                                      'description': cls_map[tag] if tag and tag in cls_map else ''}
        self.filetype = 'alto'

    def _parse_page(self):
        with open(self.filename, 'rb') as fp:
            base_directory = self.filename.parent

            try:
                doc = etree.parse(fp)
            except etree.XMLSyntaxError as e:
                raise ValueError(f'Parsing {self.filename} failed: {e}')
            image = doc.find('.//{*}Page')
            if image is None or image.get('imageFilename') is None:
                raise ValueError(f'No valid image filename found in PageXML file {self.filename}')
            try:
                self.base_dir = {'left-to-right': 'L',
                                 'right-to-left': 'R',
                                 'top-to-bottom': 'L',
                                 'bottom-to-top': 'R',
                                 None: None}[image.get('readingDirection')]
            except KeyError:
                logger.warning(f'Invalid value {image.get("readingDirection")} encountered in page-level reading direction.')
            self.imagename = base_directory.joinpath(image.get('imageFilename'))
            self.image_size = int(image.get('imageWidth')), int(image.get('imageHeight'))

            # find all image regions
            regions = [reg for reg in image.iterfind('./{*}*')]
            # parse region type and coords
            region_data = defaultdict(list)
            tr_region_order = []

            self._tag_set = set(('default',))
            tmp_transkribus_line_order = defaultdict(list)

            for region in regions:
                if not any([True if region.tag.endswith(k) else False for k in page_regions.keys()]):
                    continue
                coords = region.find('./{*}Coords')
                if coords is not None and not coords.get('points').isspace() and len(coords.get('points')):
                    try:
                        coords = self._parse_page_coords(coords.get('points'))
                    except Exception:
                        logger.warning('Region {} without coordinates'.format(region.get('id')))
                        coords = None
                else:
                    logger.warning('Region {} without coordinates'.format(region.get('id')))
                    coords = None
                rtype = region.get('type')
                # parse transkribus-style custom field if possible
                custom_str = region.get('custom')
                if custom_str:
                    cs = self._parse_page_custom(custom_str)
                    if not rtype and 'structure' in cs and 'type' in cs['structure']:
                        rtype = cs['structure']['type']
                    # transkribus-style reading order
                    if 'readingOrder' in cs and 'index' in cs['readingOrder']:
                        tr_region_order.append((region.get('id'), int(cs['readingOrder']['index'])))
                # fall back to default region type if nothing is given
                if not rtype:
                    rtype = page_regions[region.tag.split('}')[-1]]
                region_data[rtype].append(Region(id=region.get('id'), boundary=coords, tags={'type': rtype}))
                # register implicit reading order
                self._orders['region_implicit']['order'].append(region.get('id'))

                # parse line information
                for line in region.iterfind('./{*}TextLine'):
                    pol = line.find('./{*}Coords')
                    boundary = None
                    if pol is not None and not pol.get('points').isspace() and len(pol.get('points')):
                        try:
                            boundary = self._parse_page_coords(pol.get('points'))
                        except Exception:
                            logger.info('TextLine {} without polygon'.format(line.get('id')))
                    else:
                        logger.info('TextLine {} without polygon'.format(line.get('id')))
                    base = line.find('./{*}Baseline')
                    baseline = None
                    if base is not None and not base.get('points').isspace() and len(base.get('points')):
                        try:
                            baseline = self._parse_page_coords(base.get('points'))
                        except Exception:
                            logger.info('TextLine {} without baseline'.format(line.get('id')))
                            continue
                    else:
                        logger.info('TextLine {} without baseline'.format(line.get('id')))
                        continue
                    text = ''
                    manual_transcription = line.find('./{*}TextEquiv')
                    if manual_transcription is not None:
                        transcription = manual_transcription
                    else:
                        transcription = line
                    for el in transcription.findall('.//{*}Unicode'):
                        if el.text:
                            text += el.text
                    # retrieve line tags if custom string is set and contains
                    tags = {'type': 'default'}
                    split_type = None
                    custom_str = line.get('custom')
                    if custom_str:
                        cs = self._parse_page_custom(custom_str)
                        if 'structure' in cs and 'type' in cs['structure']:
                            tags['type'] = cs['structure']['type']
                            self._tag_set.add(tags['type'])
                        # retrieve data split if encoded in custom string.
                        if 'split' in cs and 'type' in cs['split'] and cs['split']['type'] in ['train', 'validation', 'test']:
                            split_type = cs['split']['type']
                            tags['split'] = split_type
                            self._tag_set.add(split_type)
                        if 'readingOrder' in cs and 'index' in cs['readingOrder']:
                            # look up region index from parent
                            reg_cus = self._parse_page_custom(line.getparent().get('custom'))
                            if 'readingOrder' not in reg_cus or 'index' not in reg_cus['readingOrder']:
                                logger.warning('Incomplete `custom` attribute reading order found.')
                            else:
                                tmp_transkribus_line_order[int(reg_cus['readingOrder']['index'])].append((int(cs['readingOrder']['index']), line.get('id')))

                    self._lines[line.get('id')] = BaselineLine(id=line.get('id'),
                                                               baseline=baseline,
                                                               boundary=boundary,
                                                               text=text,
                                                               tags=tags,
                                                               split=split_type,
                                                               regions=[region.get('id')])
                    # register implicit reading order
                    self._orders['line_implicit']['order'].append(line.get('id'))

            # add transkribus-style region order
            self._orders['region_transkribus'] = {'order': [x[0] for x in sorted(tr_region_order, key=lambda k: k[1])],
                                                  'is_total': True if len(set(map(lambda x: x[0], tr_region_order))) == len(tr_region_order) else False,
                                                  'description': 'Explicit region order from `custom` attribute'}

            self._regions = region_data

            if tmp_transkribus_line_order:
                # sort by regions
                tmp_reg_order = sorted(((k, v) for k, v in tmp_transkribus_line_order.items()), key=lambda k: k[0])
                # flatten
                tr_line_order = []
                for _, lines in tmp_reg_order:
                    tr_line_order.extend([x[1] for x in sorted(lines, key=lambda k: k[0])])
                self._orders['line_transkribus'] = {'order': tr_line_order,
                                                    'is_total': True,
                                                    'description': 'Explicit line order from `custom` attribute'}

            # parse explicit reading orders if they exist
            ro_el = doc.find('.//{*}ReadingOrder')
            if ro_el is not None:
                reading_orders = ro_el.getchildren()
                # UnorderedGroup at top-level => treated as multiple reading orders
                if len(reading_orders) == 1 and reading_orders[0].tag.endswith('UnorderedGroup'):
                    reading_orders = reading_orders.getchildren()

                def _parse_group(el):

                    _ro = []
                    if el.tag.endswith('UnorderedGroup'):
                        _ro = [_parse_group(x) for x in el.iterchildren()]
                        is_total = False  # NOQA
                    elif el.tag.endswith('OrderedGroup'):
                        _ro.extend(_parse_group(x) for x in el.iterchildren())
                    else:
                        return el.get('regionRef')
                    return _ro

                for ro in reading_orders:
                    is_total = True
                    self._orders[ro.get('id')] = {'order': _parse_group(ro),
                                                  'is_total': is_total,
                                                  'description': ro.get('caption') if ro.get('caption') else ''}

        if len(self._tag_set) > 1:
            self.has_tags = True
        else:
            self.has_tags = False

        self.filetype = 'page'

    @property
    def regions(self):
        return self._regions

    @property
    def lines(self):
        return self._lines

    @property
    def reading_orders(self):
        return self._orders

    def get_sorted_lines(self, ro='line_implicit'):
        """
        Returns ordered baselines from particular reading order.
        """
        if ro not in self.reading_orders:
            raise ValueError(f'Unknown reading order {ro}')

        def _traverse_ro(el):
            _ro = []
            if isinstance(el, list):
                _ro = [_traverse_ro(x) for x in el]
            else:
                # if line directly append to ro
                if el in self.lines:
                    return self.lines[el]
                # substitute lines if region in RO
                elif el in [reg['id'] for regs in self.regions.values() for reg in regs]:
                    _ro.extend(self.get_sorted_lines_by_region(el))
                else:
                    raise ValueError(f'Invalid reading order {ro}')
            return _ro

        _ro = self.reading_orders[ro]
        return _traverse_ro(_ro['order'])

    def get_sorted_regions(self, ro='region_implicit'):
        """
        Returns ordered regions from particular reading order.
        """
        if ro not in self.reading_orders:
            raise ValueError(f'Unknown reading order {ro}')

        regions = {reg.id: key for key, regs in self.regions.items() for reg in regs}

        def _traverse_ro(el):
            _ro = []
            if isinstance(el, list):
                _ro = [_traverse_ro(x) for x in el]
            else:
                # if region directly append to ro
                if el in regions.keys():
                    return [reg for reg in self.regions[regions[el]] if reg.id == el][0]
                else:
                    raise ValueError(f'Invalid reading order {ro}')
            return _ro

        _ro = self.reading_orders[ro]
        return _traverse_ro(_ro['order'])

    def get_sorted_lines_by_region(self, region, ro='line_implicit'):
        """
        Returns ordered lines in region.
        """
        if ro not in self.reading_orders:
            raise ValueError(f'Unknown reading order {ro}')
        if self.reading_orders[ro]['is_total'] is False:
            raise ValueError('Fetching lines by region of a non-total order is not supported')
        lines = [(id, line) for id, line in self._lines.items() if line.regions[0] == region]
        for line in lines:
            if line[0] not in self.reading_orders[ro]['order']:
                raise ValueError('Fetching lines by region is only possible for flat orders')
        return sorted(lines, key=lambda k: self.reading_orders[ro]['order'].index(k[0]))

    def get_lines_by_tag(self, key, value):
        return {k: v for k, v in self._lines.items() if v.tags.get(key) == value}

    def get_lines_by_split(self, split: Literal['train', 'validation', 'test']):
        return {k: v for k, v in self._lines.items() if v.tags.get('split') == split}

    @property
    def tags(self):
        return self._tag_set

    @property
    def splits(self):
        return self._split_set

    @staticmethod
    def _parse_alto_pointstype(coords: str) -> Sequence[Tuple[float, float]]:
        """
        ALTO's PointsType is underspecified so a variety of serializations are valid:

            x0, y0 x1, y1 ...
            x0 y0 x1 y1 ...
            (x0, y0) (x1, y1) ...
            (x0 y0) (x1 y1) ...

        Returns:
            A list of tuples [(x0, y0), (x1, y1), ...]
        """
        float_re = re.compile(r'[-+]?(\d+(\.\d*)?|\.\d+)([eE][-+]?\d+)?')
        points = [float(point.group()) for point in float_re.finditer(coords)]
        if len(points) % 2:
            raise ValueError(f'Odd number of points in points sequence: {points}')
        pts = zip(points[::2], points[1::2])
        return [k for k, g in groupby(pts)]

    @staticmethod
    def _parse_page_custom(s):
        o = {}
        s = s.strip()
        l_chunks = [l_chunk for l_chunk in s.split('}') if l_chunk.strip()]
        if l_chunks:
            for chunk in l_chunks:
                tag, vals = chunk.split('{')
                tag_vals = {}
                vals = [val.strip() for val in vals.split(';') if val.strip()]
                for val in vals:
                    key, *val = val.split(':')
                    tag_vals[key] = ":".join(val)
                o[tag.strip()] = tag_vals
        return o

    @staticmethod
    def _parse_page_coords(coords):
        points = [x for x in coords.split(' ')]
        points = [int(c) for point in points for c in point.split(',')]
        pts = zip(points[::2], points[1::2])
        return [k for k, g in groupby(pts)]

    def __str__(self):
        return f'XMLPage {self.filename} (format: {self.filetype}, image: {self.imagename})'

    def __repr__(self):
        return f'XMLPage(filename={self.filename}, filetype={self.filetype})'

    def to_container(self) -> Segmentation:
        """
        Returns a Segmentation object.
        """
        return Segmentation(type='baselines',
                            imagename=self.imagename,
                            text_direction='horizontal_lr',
                            script_detection=True,
                            lines=self.get_sorted_lines(),
                            regions=self._regions,
                            line_orders=[])



# -*- coding: utf-8 -*-
import unittest

from pytest import raises
from torch import IntTensor

from kraken.lib import codec
from kraken.lib.exceptions import KrakenCodecException, KrakenEncodeException


class TestCodec(unittest.TestCase):

    """
    Testing codec mapping routines
    """

    def setUp(self):
        # codec mapping one code point to one label
        self.o2o_codec = codec.PytorchCodec('ab')
        self.o2o_codec_strict = codec.PytorchCodec('ab', strict=True)
        # codec mapping many code points to one label
        self.m2o_codec = codec.PytorchCodec(['aaa' , 'aa', 'a', 'b'])
        self.m2o_codec_strict = codec.PytorchCodec(['aaa' , 'aa', 'a', 'b'], strict=True)
        # codec mapping one code point to many labels
        self.o2m_codec = codec.PytorchCodec({'a': [10, 11, 12], 'b': [12, 45, 80]})
        self.o2m_codec_strict = codec.PytorchCodec({'a': [10, 11, 12], 'b': [12, 45, 80]}, strict=True)
        # codec mapping many code points to many labels
        self.m2m_codec = codec.PytorchCodec({'aaa': [10, 11, 12], 'aa': [9, 9], 'a': [11], 'bb': [15], 'b': [12]})
        self.m2m_codec_strict = codec.PytorchCodec({'aaa': [10, 11, 12], 'aa': [9, 9], 'a': [11], 'bb': [15], 'b': [12]}, strict=True)

        self.invalid_c_sequence = 'aaababbcaaa'
        self.valid_c_sequence = 'aaababbaaabbbb'

        self.ada_sequence = 'cdaabae'

        self.invalid_l_sequence = [(45, 78, 778, 0.3793492615638364),
                                   (10, 203, 859, 0.9485075253700872),
                                   (11, 70, 601, 0.7885297329523855),
                                   (12, 251, 831, 0.7216817042926938),
                                   (900, 72, 950, 0.27609823017048707)]

    def test_o2o_encode(self):
        """
        Test correct encoding of one-to-one code point sequence
        """
        self.assertTrue(self.o2o_codec.encode(self.valid_c_sequence).eq(
                        IntTensor([1, 1, 1, 2, 1, 2, 2, 1, 1, 1, 2, 2, 2, 2])).all())

    def test_m2o_encode(self):
        """
        Test correct encoding of many-to-one code point sequence
        """
        self.assertTrue(self.m2o_codec.encode(self.valid_c_sequence).eq(
                        IntTensor([3, 4, 1, 4, 4, 3, 4, 4, 4, 4])).all())

    def test_o2m_encode(self):
        """
        Test correct encoding of one-to-many code point sequence
        """
        self.assertTrue(self.o2m_codec.encode(self.valid_c_sequence).eq(
                        IntTensor([10, 11, 12, 10, 11, 12, 10, 11, 12,
                                   12, 45, 80, 10, 11, 12, 12, 45, 80, 12, 45,
                                   80, 10, 11, 12, 10, 11, 12, 10, 11, 12, 12,
                                   45, 80, 12, 45, 80, 12, 45, 80, 12, 45,
                                   80])).all())

    def test_m2m_encode(self):
        """
        Test correct encoding of many-to-many code point sequence
        """
        self.assertTrue(self.m2m_codec.encode(self.valid_c_sequence).eq(
                        IntTensor([10, 11, 12, 12, 11, 15, 10, 11, 12, 15, 15])).all())

    def test_o2o_decode(self):
        """
        Test correct decoding of one-to-one label sequence
        """
        self.assertEqual(''.join(x[0] for x in self.o2o_codec.decode([(1, 288, 652, 0.8537325587315542),
                                                                      (1, 120, 861, 0.4968470297302481),
                                                                      (1, 372, 629, 0.008650773294205938),
                                                                      (2, 406, 831, 0.15637985875540783),
                                                                      (1, 3, 824, 0.26475146828232776),
                                                                      (2, 228, 959, 0.3062689368044844),
                                                                      (2, 472, 679, 0.8677848554329698),
                                                                      (1, 482, 771, 0.6055591197109657),
                                                                      (1, 452, 606, 0.40744265053745055),
                                                                      (1, 166, 879, 0.7509269177978337),
                                                                      (2, 92, 729, 0.34554103785480306),
                                                                      (2, 227, 959, 0.3006394689033981),
                                                                      (2, 341, 699, 0.07798704843315862),
                                                                      (2, 142, 513, 0.9933850573241767)])),
                         'aaababbaaabbbb')

    def test_m2o_decode(self):
        """
        Test correct decoding of many-to-one label sequence
        """
        self.assertEqual(''.join(x[0] for x in self.m2o_codec.decode([(3, 28, 967, 0.07761440833942468),
                                                                      (4, 282, 565, 0.4946281412618093),
                                                                      (1, 411, 853, 0.7767301050586806),
                                                                      (4, 409, 501, 0.47915609540996495),
                                                                      (4, 299, 637, 0.7755889399450564),
                                                                      (3, 340, 834, 0.726656062406549),
                                                                      (4, 296, 846, 0.2274859668684881),
                                                                      (4, 238, 695, 0.32982930128257815),
                                                                      (4, 187, 970, 0.43354272748701805),
                                                                      (4, 376, 863, 0.24483897879550764)])),
                         'aaababbaaabbbb')

    def test_o2m_decode(self):
        """
        Test correct decoding of one-to-many label sequence
        """
        self.assertEqual(''.join(x[0] for x in self.o2m_codec.decode([(10, 35, 959, 0.43819571289990644),
                                                                      (11, 361, 904, 0.1801115018592916),
                                                                      (12, 15, 616, 0.5987506334315549),
                                                                      (10, 226, 577, 0.6178248939780698),
                                                                      (11, 227, 814, 0.31531097360327787),
                                                                      (12, 390, 826, 0.7706594984014595),
                                                                      (10, 251, 579, 0.9442530315305507),
                                                                      (11, 269, 870, 0.4475979925584944),
                                                                      (12, 456, 609, 0.9396137478409995),
                                                                      (12, 60, 757, 0.06416607235266458),
                                                                      (45, 318, 918, 0.8129458423341515),
                                                                      (80, 15, 914, 0.49773432435726517),
                                                                      (10, 211, 648, 0.7919220961861382),
                                                                      (11, 326, 804, 0.7852387442556333),
                                                                      (12, 93, 978, 0.9376801123379804),
                                                                      (12, 23, 698, 0.915543635886972),
                                                                      (45, 71, 599, 0.8137750423628737),
                                                                      (80, 167, 980, 0.6501035181890226),
                                                                      (12, 259, 823, 0.3122860659712233),
                                                                      (45, 312, 948, 0.20582589628806058),
                                                                      (80, 430, 694, 0.3528792552966924),
                                                                      (10, 470, 866, 0.0685524032330419),
                                                                      (11, 459, 826, 0.39354887700146846),
                                                                      (12, 392, 926, 0.4102018609185847),
                                                                      (10, 271, 592, 0.1877915301623876),
                                                                      (11, 206, 995, 0.21614062190981576),
                                                                      (12, 466, 648, 0.3106914763314057),
                                                                      (10, 368, 848, 0.28715379701274113),
                                                                      (11, 252, 962, 0.5535299604896257),
                                                                      (12, 387, 709, 0.844810014550603),
                                                                      (12, 156, 916, 0.9803695305965802),
                                                                      (45, 150, 555, 0.5969071330809561),
                                                                      (80, 381, 922, 0.5608300913697513),
                                                                      (12, 35, 762, 0.5227506455088722),
                                                                      (45, 364, 931, 0.7205481732247938),
                                                                      (80, 341, 580, 0.536934566913969),
                                                                      (12, 79, 919, 0.5136066153481802),
                                                                      (45, 377, 773, 0.6507467790760987),
                                                                      (80, 497, 931, 0.7635100185309783),
                                                                      (12, 76, 580, 0.9542477438586341),
                                                                      (45, 37, 904, 0.4299813924853797),
                                                                      (80, 425, 638, 0.6825047210425983)])),
                         'aaababbaaabbbb')

    def test_m2m_decode(self):
        """
        Test correct decoding of many-to-many label sequence
        """
        self.assertEqual(''.join(x[0] for x in self.m2m_codec.decode([(10, 313, 788, 0.9379917930525369),
                                                                      (11, 117, 793, 0.9974374577004185),
                                                                      (12, 50, 707, 0.020074164253385374),
                                                                      (12, 382, 669, 0.525910770170754),
                                                                      (10, 458, 833, 0.4292373233167248),
                                                                      (15, 45, 831, 0.5759709886686226),
                                                                      (10, 465, 729, 0.8492104897235935),
                                                                      (11, 78, 800, 0.24733538459309445),
                                                                      (12, 375, 872, 0.26908722769105353),
                                                                      (15, 296, 889, 0.44251812620463726),
                                                                      (15, 237, 930, 0.5456105208117391)])),
                         'aaabbbaaabbbb')

    def test_o2o_decode_invalid_nonstrict(self):
        """
        Test correct handling of undecodable sequences (one-to-one decoder)
        """
        self.assertEqual(self.o2o_codec.decode(self.invalid_l_sequence), [])

    def test_m2o_decode_invalid_nonstrict(self):
        """
        Test correct handling of undecodable sequences (many-to-one decoder)
        """
        self.assertEqual(self.m2o_codec.decode(self.invalid_l_sequence), [])

    def test_o2m_decode_invalid_nonstrict(self):
        """
        Test correct handling of undecodable sequences (one-to-many decoder)
        """
        self.assertEqual(self.o2m_codec.decode(self.invalid_l_sequence),
                         [('a', 203, 831, 0.8195729875383888)])

    def test_m2m_decode_invalid_nonstrict(self):
        """
        Test correct handling of undecodable sequences (many-to-many decoder)
        """
        self.assertEqual(self.m2m_codec.decode(self.invalid_l_sequence),
                         [('a', 203, 831, 0.8195729875383888),
                          ('a', 203, 831, 0.8195729875383888),
                          ('a', 203, 831, 0.8195729875383888)])


    def test_o2o_encode_invalid_nonstrict(self):
        """
        Test correct handling of noisy character sequences (one-to-one encoder)
        """
        self.assertTrue(self.o2o_codec.encode(self.invalid_c_sequence).eq(
                        IntTensor([1, 1, 1, 2, 1, 2, 2, 1, 1, 1])).all())

    def test_m2o_encode_invalid_nonstrict(self):
        """
        Test correct handling of noisy character sequences (many-to-one encoder)
        """
        self.assertTrue(self.m2o_codec.encode(self.invalid_c_sequence).eq(
                        IntTensor([3, 4, 1, 4, 4, 3])).all())

    def test_o2m_encode_invalid_nonstrict(self):
        """
        Test correct handling of noisy character sequences (one-to-many encoder)
        """
        self.assertTrue(self.o2m_codec.encode(self.invalid_c_sequence).eq(
                        IntTensor([10, 11, 12, 10, 11, 12, 10, 11, 12, 12, 45,
                                   80, 10, 11, 12, 12, 45, 80, 12, 45, 80, 10,
                                   11, 12, 10, 11, 12, 10, 11, 12])).all())

    def test_m2m_encode_invalid_nonstrict(self):
        """
        Test correct handling of noisy character sequences (many-to-many encoder)
        """
        self.assertTrue(self.m2m_codec.encode(self.invalid_c_sequence).eq(
                        IntTensor([10, 11, 12, 12, 11, 15, 10, 11, 12])).all())

    def test_o2o_decode_invalid(self):
        """
        Test correct handling of undecodable sequences (one-to-one decoder) in strict mode
        """
        with raises(KrakenEncodeException):
            self.o2o_codec_strict.decode(self.invalid_l_sequence)

    def test_m2o_decode_invalid(self):
        """
        Test correct handling of undecodable sequences (many-to-one decoder) in strict mode
        """
        with raises(KrakenEncodeException):
            self.m2o_codec_strict.decode(self.invalid_l_sequence)

    def test_o2m_decode_invalid(self):
        """
        Test correct handling of undecodable sequences (one-to-many decoder) in strict mode
        """
        with raises(KrakenEncodeException):
            self.o2m_codec_strict.decode(self.invalid_l_sequence)

    def test_m2m_decode_invalid(self):
        """
        Test correct handling of undecodable sequences (many-to-many decoder) in strict mode
        """
        with raises(KrakenEncodeException):
            self.m2m_codec_strict.decode(self.invalid_l_sequence)

    def test_o2o_encode_invalid(self):
        """
        Test correct handling of unencodable sequences (one-to-one encoder) in strict mode
        """
        with raises(KrakenEncodeException):
            self.o2o_codec_strict.encode(self.invalid_c_sequence)

    def test_m2o_encode_invalid(self):
        """
        Test correct handling of unencodable sequences (many-to-one encoder) in strict mode
        """
        with raises(KrakenEncodeException):
            self.m2o_codec_strict.encode(self.invalid_c_sequence)

    def test_o2m_encode_invalid(self):
        """
        Test correct handling of unencodable sequences (one-to-many encoder) in strict mode
        """
        with raises(KrakenEncodeException):
            self.o2m_codec_strict.encode(self.invalid_c_sequence)

    def test_m2m_encode_invalid(self):
        """
        Test correct handling of unencodable sequences (many-to-many encoder) in strict mode
        """
        with raises(KrakenEncodeException):
            self.m2m_codec_strict.encode(self.invalid_c_sequence)

    def test_codec_add_simple(self):
        """
        Test adding of new code points to codec.
        """
        prev_len = len(self.o2o_codec)
        codec = self.o2o_codec.add_labels('cde')
        self.assertEqual(len(codec), prev_len + 3)
        self.assertTrue(codec.encode(self.ada_sequence).eq(
                        IntTensor([3, 4, 1, 1, 2, 1, 5])).all())

    def test_codec_add_list(self):
        """
        Test adding of new code points to codec.
        """
        prev_len = len(self.o2o_codec)
        codec = self.o2o_codec.add_labels(['cd', 'e'])
        self.assertEqual(len(codec), prev_len + 2)
        self.assertTrue(codec.encode(self.ada_sequence).eq(
                        IntTensor([3, 1, 1, 2, 1, 4])).all())

    def test_codec_add_dict(self):
        """
        Test adding of new code points to codec.
        """
        prev_len = len(self.o2o_codec)
        codec = self.o2o_codec.add_labels({'cd': [3], 'e': [4]})
        self.assertEqual(len(codec), prev_len + 2)
        self.assertTrue(codec.encode(self.ada_sequence).eq(
                        IntTensor([3, 1, 1, 2, 1, 4])).all())

    def test_codec_merge_both(self):
        """
        Test merging of a codec adding and removing code points
        """
        merge_codec = codec.PytorchCodec('acde')
        new_codec, del_labels = self.o2o_codec.merge(merge_codec)
        self.assertEqual(del_labels, {2})
        self.assertEqual(new_codec.c2l, {'a': [1], 'c': [2], 'd': [3], 'e': [4]})

    def test_codec_merge_add(self):
        """
        Test merging of a codec adding and removing code points
        """
        merge_codec = codec.PytorchCodec('abcde')
        new_codec, del_labels = self.o2o_codec.merge(merge_codec)
        self.assertEqual(del_labels, set())
        self.assertEqual(new_codec.c2l, {'a': [1], 'b': [2], 'c': [3], 'd': [4], 'e': [5]})

    def test_codec_merge_remove(self):
        """
        Test merging of a codec removing code points
        """
        merge_codec = codec.PytorchCodec('a')
        new_codec, del_labels = self.o2o_codec.merge(merge_codec)
        self.assertEqual(del_labels, {2})
        self.assertEqual(new_codec.c2l, {'a': [1]})



#
# Copyright 2022 Benjamin Kiessling
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
"""
kraken.ketos.linegen
~~~~~~~~~~~~~~~~~~~~

Command line driver for synthetic recognition training data generation.
"""
import click


@click.command('linegen', deprecated=True)
@click.pass_context
@click.option('-f', '--font', default='sans',
              help='Font family to render texts in.')
@click.option('-n', '--maxlines', type=click.INT, default=0,
              help='Maximum number of lines to generate')
@click.option('-e', '--encoding', default='utf-8',
              help='Decode text files with given codec.')
@click.option('-u', '--normalization',
              type=click.Choice(['NFD', 'NFKD', 'NFC', 'NFKC']), default=None,
              help='Normalize ground truth')
@click.option('-ur', '--renormalize',
              type=click.Choice(['NFD', 'NFKD', 'NFC', 'NFKC']), default=None,
              help='Renormalize text for rendering purposes.')
@click.option('--reorder/--no-reorder', default=False, help='Reorder code points to display order')
@click.option('-fs', '--font-size', type=click.INT, default=32,
              help='Font size to render texts in.')
@click.option('-fw', '--font-weight', type=click.INT, default=400,
              help='Font weight to render texts in.')
@click.option('-l', '--language',
              help='RFC-3066 language tag for language-dependent font shaping')
@click.option('-ll', '--max-length', type=click.INT, default=None,
              help="Discard lines above length (in Unicode codepoints).")
@click.option('--strip/--no-strip', help="Remove whitespace from start and end "
              "of lines.")
@click.option('-D', '--disable-degradation', is_flag=True, help='Dont degrade '
              'output lines.')
@click.option('-a', '--alpha', type=click.FLOAT, default=1.5,
              help="Mean of folded normal distribution for sampling foreground pixel flip probability")
@click.option('-b', '--beta', type=click.FLOAT, default=1.5,
              help="Mean of folded normal distribution for sampling background pixel flip probability")
@click.option('-d', '--distort', type=click.FLOAT, default=1.0,
              help='Mean of folded normal distribution to take distortion values from')
@click.option('-ds', '--distortion-sigma', type=click.FLOAT, default=20.0,
              help='Mean of folded normal distribution to take standard deviations for the '
              'Gaussian kernel from')
@click.option('--legacy/--no-legacy', default=False,
              help='Use ocropy-style degradations')
@click.option('-o', '--output', type=click.Path(), default='training_data',
              help='Output directory')
@click.argument('text', nargs=-1, type=click.Path(exists=True))
def line_generator(ctx, font, maxlines, encoding, normalization, renormalize,
                   reorder, font_size, font_weight, language, max_length, strip,
                   disable_degradation, alpha, beta, distort, distortion_sigma,
                   legacy, output, text):
    """
    Generates artificial text line training data.
    """
    import errno
    import logging
    import os
    import unicodedata
    from typing import Set

    import numpy as np
    from bidi.algorithm import get_display

    from kraken import linegen
    from kraken.lib.exceptions import KrakenCairoSurfaceException
    from kraken.lib.progress import KrakenProgressBar
    from kraken.lib.util import make_printable

    from .util import message

    logging.captureWarnings(True)
    logger = logging.getLogger('kraken')

    lines: Set[str] = set()
    if not text:
        return
    with KrakenProgressBar() as progress:
        read_task = progress.add_task('Reading texts', total=len(text), visible=True if not ctx.meta['verbose'] else False)
        for t in text:
            with click.open_file(t, encoding=encoding) as fp:
                logger.info('Reading {}'.format(t))
                for line in fp:
                    lines.add(line.rstrip('\r\n'))
            progress.update(read_task, advance=1)

    if normalization:
        lines = set([unicodedata.normalize(normalization, line) for line in lines])
    if strip:
        lines = set([line.strip() for line in lines])
    if max_length:
        lines = set([line for line in lines if len(line) < max_length])
    logger.info('Read {} lines'.format(len(lines)))
    message('Read {} unique lines'.format(len(lines)))
    if maxlines and maxlines < len(lines):
        message('Sampling {} lines\t'.format(maxlines), nl=False)
        llist = list(lines)
        lines = set(llist[idx] for idx in np.random.randint(0, len(llist), maxlines))
        message('\u2713', fg='green')
    try:
        os.makedirs(output)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise

    # calculate the alphabet and print it for verification purposes
    alphabet: Set[str] = set()
    for line in lines:
        alphabet.update(line)
    chars = []
    combining = []
    for char in sorted(alphabet):
        k = make_printable(char)
        if k != char:
            combining.append(k)
        else:
            chars.append(k)
    message('Σ (len: {})'.format(len(alphabet)))
    message('Symbols: {}'.format(''.join(chars)))
    if combining:
        message('Combining Characters: {}'.format(', '.join(combining)))
    lg = linegen.LineGenerator(font, font_size, font_weight, language)
    with KrakenProgressBar() as progress:
        gen_task = progress.add_task('Writing images', total=len(lines), visible=True if not ctx.meta['verbose'] else False)
        for idx, line in enumerate(lines):
            logger.info(line)
            try:
                if renormalize:
                    im = lg.render_line(unicodedata.normalize(renormalize, line))
                else:
                    im = lg.render_line(line)
            except KrakenCairoSurfaceException as e:
                logger.info('{}: {} {}'.format(e.message, e.width, e.height))
                continue
            if not disable_degradation and not legacy:
                im = linegen.degrade_line(im, alpha=alpha, beta=beta)
                im = linegen.distort_line(im, abs(np.random.normal(distort)), abs(np.random.normal(distortion_sigma)))
            elif legacy:
                im = linegen.ocropy_degrade(im)
            im.save('{}/{:06d}.png'.format(output, idx))
            with open('{}/{:06d}.gt.txt'.format(output, idx), 'wb') as fp:
                if reorder:
                    fp.write(get_display(line).encode('utf-8'))
                else:
                    fp.write(line.encode('utf-8'))
            progress.update(gen_task, advance=1)



#
# Copyright 2019 Benjamin Kiessling
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
"""
Processing for baseline segmenter output
"""
import logging
from collections import defaultdict
from typing import (TYPE_CHECKING, Dict, List, Literal, Optional, Sequence,
                    Tuple, Union, TypeVar, Any, Generator)

import numpy as np
import shapely.geometry as geom
import torch
import torch.nn.functional as F
from PIL import Image, ImageDraw
from PIL.Image import Resampling, Transform
from scipy.ndimage import (binary_erosion, distance_transform_cdt,
                           gaussian_filter, maximum_filter, affine_transform)
from scipy.signal import convolve2d
from scipy.spatial.distance import pdist, squareform
from shapely.ops import nearest_points, unary_union
from shapely.validation import explain_validity
from skimage import draw, filters
from skimage.filters import sobel
from skimage.graph import MCP_Connect
from skimage.measure import (approximate_polygon, label, regionprops,
                             subdivide_polygon)
from skimage.morphology import skeletonize
from skimage.transform import (AffineTransform, PiecewiseAffineTransform, warp)

from kraken.lib import default_specs
from kraken.lib.exceptions import KrakenInputException

if TYPE_CHECKING:
    from kraken.containers import Segmentation, BBoxLine, BaselineLine
    from kraken.lib.vgsl import TorchVGSLModel


_T_pil_or_np = TypeVar('_T_pil_or_np', Image.Image, np.ndarray)

logger = logging.getLogger('kraken')

__all__ = ['reading_order',
           'neural_reading_order',
           'vectorize_lines',
           'calculate_polygonal_environment',
           'polygonal_reading_order',
           'scale_polygonal_lines',
           'scale_regions',
           'compute_polygon_section',
           'extract_polygons']


# faster implementation of PiecewiseAffineTransform - see
# https://github.com/scikit-image/scikit-image/issues/6864 and
# https://github.com/scikit-image/scikit-image/pull/6963
class FastPiecewiseAffineTransform(PiecewiseAffineTransform):
    def __call__(self, coords):
        coords = np.asarray(coords)

        simplex = self._tesselation.find_simplex(coords)

        affines = np.array(
            [self.affines[i].params for i in range(len(self._tesselation.simplices))]
        )[simplex]

        pts = np.c_[coords, np.ones((coords.shape[0], 1))]

        result = np.einsum("ij,ikj->ik", pts, affines)
        result[simplex == -1, :] = -1

        return result


def reading_order(lines: Sequence[Tuple[slice, slice]], text_direction: Literal['lr', 'rl'] = 'lr') -> np.ndarray:
    """Given the list of lines (a list of 2D slices), computes
    the partial reading order.  The output is a binary 2D array
    such that order[i,j] is true if line i comes before line j
    in reading order."""

    logger.info('Compute reading order on {} lines in {} direction'.format(len(lines), text_direction))

    order = np.zeros((len(lines), len(lines)), 'B')

    def _x_overlaps(u, v):
        return u[1].start < v[1].stop and u[1].stop > v[1].start

    def _above(u, v):
        return u[0].start < v[0].start

    def _left_of(u, v):
        return u[1].stop < v[1].start

    def _separates(w, u, v):
        if w == u or w == v:
            return 0
        if w[0].stop < min(u[0].start, v[0].start):
            return 0
        if w[0].start > max(u[0].stop, v[0].stop):
            return 0
        if w[1].start < u[1].stop and w[1].stop > v[1].start:
            return 1
        return 0

    if text_direction == 'rl':
        def horizontal_order(u, v):
            return not _left_of(u, v)
    else:
        horizontal_order = _left_of

    for i, u in enumerate(lines):
        for j, v in enumerate(lines):
            if _x_overlaps(u, v):
                if _above(u, v):
                    order[i, j] = 1
            else:
                if [w for w in lines if _separates(w, u, v)] == []:
                    if horizontal_order(u, v):
                        order[i, j] = 1
    return order


def topsort(order: np.ndarray) -> List[int]:
    """Given a binary array defining a partial order (o[i,j]==True means i<j),
    compute a topological sort.  This is a quick and dirty implementation
    that works for up to a few thousand elements."""
    logger.info('Perform topological sort on partially ordered lines')
    n = len(order)
    visited = np.zeros(n)
    L = []

    def _visit(k):
        if visited[k]:
            return
        visited[k] = 1
        a, = np.nonzero(np.ravel(order[:, k]))
        for line in a:
            _visit(line)
        L.append(k)

    for k in range(n):
        _visit(k)
    return L


def moore_neighborhood(current, backtrack):
    operations = np.array([[-1, 0], [-1, 1], [0, 1], [1, 1], [1, 0], [1, -1],
                           [0, -1], [-1, -1]])
    neighbors = (current + operations).astype(int)

    for i, point in enumerate(neighbors):
        if np.all(point == backtrack):
            # we return the sorted neighborhood
            return np.concatenate((neighbors[i:], neighbors[:i]))
    return 0


def boundary_tracing(region):
    """
    Find coordinates of the region's boundary. The region must not have isolated
    points.

    Code copied from
    https://github.com/machine-shop/deepwings/blob/master/deepwings/method_features_extraction/image_processing.py#L185

    Args:
        region: object obtained with skimage.measure.regionprops().

    Returns:
        List of coordinates of pixels in the boundary.
    """

    # creating the binary image
    coords = region.coords
    maxs = np.amax(coords, axis=0)
    binary = np.zeros((maxs[0] + 2, maxs[1] + 2))
    x = coords[:, 1]
    y = coords[:, 0]
    binary[tuple([y, x])] = 1

    # initialization
    # starting point is the most upper left point
    idx_start = 0
    while True:  # asserting that the starting point is not isolated
        start = [y[idx_start], x[idx_start]]
        focus_start = binary[start[0]-1:start[0]+2, start[1]-1:start[1]+2]
        if np.sum(focus_start) > 1:
            break
        idx_start += 1

    # Determining backtrack pixel for the first element
    if (binary[start[0] + 1, start[1]] == 0 and
            binary[start[0]+1, start[1]-1] == 0):
        backtrack_start = [start[0]+1, start[1]]
    else:
        backtrack_start = [start[0], start[1] - 1]

    current = start
    backtrack = backtrack_start
    boundary = []
    counter = 0

    while True:
        neighbors_current = moore_neighborhood(current, backtrack)
        y = neighbors_current[:, 0]
        x = neighbors_current[:, 1]
        idx = np.argmax(binary[tuple([y, x])])
        boundary.append(current)
        backtrack = neighbors_current[idx-1]
        current = neighbors_current[idx]
        counter += 1

        if (np.all(current == start) and np.all(backtrack == backtrack_start)):
            break

    return np.array(boundary)


def _extend_boundaries(baselines, bin_bl_map):
    # find baseline blob boundaries
    labelled = label(bin_bl_map)
    boundaries = []
    for x in regionprops(labelled):
        try:
            # skip lines with very small bounding polygons
            if x.area < 6:
                logger.info(f'Skipping baseline extension for very small blob of area {x.area}')
                continue
            b = boundary_tracing(x)
            if len(b) > 3:
                boundaries.append(geom.Polygon(b).simplify(0.01).buffer(0))
        except Exception as e:
            logger.warning(f'Boundary tracing failed in baseline elongation: {e}')
            continue

    # extend lines to polygon boundary
    for bl in baselines:
        ls = geom.LineString(bl)
        try:
            boundary_pol = next(filter(lambda x: x.contains(ls), boundaries))
        except Exception:
            continue
        # 'left' side
        if boundary_pol.contains(geom.Point(bl[0])):
            l_point = boundary_pol.boundary.intersection(geom.LineString([(bl[0][0]-10*(bl[1][0]-bl[0][0]),
                                                                           bl[0][1]-10*(bl[1][1]-bl[0][1])), bl[0]]))
            if l_point.geom_type != 'Point':
                bl[0] = np.array(nearest_points(geom.Point(bl[0]), boundary_pol)[1].coords[0], 'int').tolist()
            else:
                bl[0] = np.array(l_point.coords[0], 'int').tolist()
        # 'right' side
        if boundary_pol.contains(geom.Point(bl[-1])):
            r_point = boundary_pol.boundary.intersection(geom.LineString([(bl[-1][0]-10*(bl[-2][0]-bl[-1][0]),
                                                                           bl[-1][1]-10*(bl[-2][1]-bl[-1][1])), bl[-1]]))
            if r_point.geom_type != 'Point':
                bl[-1] = np.array(nearest_points(geom.Point(bl[-1]), boundary_pol)[1].coords[0], 'int').tolist()
            else:
                bl[-1] = np.array(r_point.coords[0], 'int').tolist()
    return baselines


class LineMCP(MCP_Connect):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.connections = dict()
        self.scores = defaultdict(lambda: np.inf)

    def create_connection(self, id1, id2, pos1, pos2, cost1, cost2):
        k = (min(id1, id2), max(id1, id2))
        s = cost1 + cost2
        if self.scores[k] > s:
            self.connections[k] = (pos1, pos2, s)
            self.scores[k] = s

    def get_connections(self):
        results = []
        for k, (pos1, pos2, s) in self.connections.items():
            results.append(np.concatenate([self.traceback(pos1), self.traceback(pos2)[::-1]]))
        return results

    def goal_reached(self, int_index, float_cumcost):
        return 2 if float_cumcost else 0


def vectorize_lines(im: np.ndarray, threshold: float = 0.17, min_length=5,
                    text_direction: str = 'horizontal'):
    """
    Vectorizes lines from a binarized array.

    Args:
        im (np.ndarray): Array of shape (3, H, W) with the first dimension
                         being probabilities for (start_separators,
                         end_separators, baseline).
        threshold (float): Threshold for baseline blob detection.
        min_length (int): Minimal length of output baselines.
        text_direction (str): Base orientation of the text line (horizontal or
                              vertical).

    Returns:
        [[x0, y0, ... xn, yn], [xm, ym, ..., xk, yk], ... ]
        A list of lists containing the points of all baseline polylines.
    """
    if text_direction not in ['horizontal', 'vertical']:
        raise ValueError(f'Invalid text direction "{text_direction}"')

    # split into baseline and separator map
    st_map = im[0]
    end_map = im[1]
    bl_map = im[2]
    bl_map = filters.sato(bl_map, black_ridges=False, mode='constant')
    bin_bl_map = bl_map > threshold
    # skeletonize
    line_skel = skeletonize(bin_bl_map)
    # find end points
    kernel = np.array([[1, 1, 1], [1, 10, 1], [1, 1, 1]])
    line_extrema = np.transpose(np.where((convolve2d(line_skel, kernel, mode='same') == 11) * line_skel))

    mcp = LineMCP(~line_skel)
    try:
        mcp.find_costs(line_extrema)
    except ValueError:
        return []

    lines = [approximate_polygon(line, 3).tolist() for line in mcp.get_connections()]
    # extend baselines to blob boundary
    lines = _extend_boundaries(lines, bin_bl_map)

    # orient lines
    f_st_map = maximum_filter(st_map, size=20)
    f_end_map = maximum_filter(end_map, size=20)

    oriented_lines = []
    for bl in lines:
        l_end = tuple(bl[0])
        r_end = tuple(bl[-1])
        if f_st_map[l_end] - f_end_map[l_end] > 0.2 and f_st_map[r_end] - f_end_map[r_end] < -0.2:
            pass
        elif f_st_map[l_end] - f_end_map[l_end] < -0.2 and f_st_map[r_end] - f_end_map[r_end] > 0.2:
            bl = bl[::-1]
        else:
            if text_direction == 'horizontal':
                logger.debug('Insufficient marker confidences in output. Defaulting to horizontal upright line.')
                if bl[0][1] > bl[-1][1]:
                    bl = bl[::-1]
            else:
                logger.debug('Insufficient marker confidences in output. Defaulting to top-to-bottom line.')
                if bl[0][0] > bl[-1][0]:
                    bl = bl[::-1]
        if geom.LineString(bl).length >= min_length:
            oriented_lines.append([x[::-1] for x in bl])
    return oriented_lines


def vectorize_regions(im: np.ndarray, threshold: float = 0.5):
    """
    Vectorizes lines from a binarized array.

    Args:
        im (np.ndarray): Array of shape (H, W) with the first dimension
                         being a probability distribution over the region.
        threshold (float): Threshold for binarization

    Returns:
        [[x0, y0, ... xn, yn], [xm, ym, ..., xk, yk], ... ]
        A list of lists containing the region polygons.
    """
    bin = im > threshold
    labelled = label(bin)
    boundaries = []
    for x in regionprops(labelled):
        boundary = boundary_tracing(x)
        if len(boundary) > 2:
            boundaries.append(geom.Polygon(boundary))
    # merge regions that overlap
    boundaries = unary_union(boundaries)
    # simplify them afterwards
    if boundaries.geom_type == 'Polygon':
        boundaries = [boundaries.boundary.simplify(10)]
    else:
        boundaries = [x.boundary.simplify(10) for x in boundaries.geoms]
    return [np.array(x.coords, dtype=np.uint)[:, [1, 0]].tolist() for x in boundaries]


def _rotate(image: _T_pil_or_np,
            angle: float,
            center: Any,
            scale: float,
            cval: int = 0,
            order: int = 0,
            use_skimage_warp: bool = False) -> Tuple[AffineTransform, _T_pil_or_np]:
    """
    Rotate an image at an angle with optional scaling

    Args:
        image (PIL.Image.Image or (H, W, C) np.ndarray): Input image
        angle (float): Angle in radians
        center (tuple): unused
        scale (float): x-Axis scaling factor
        cval (int): Padding value
        order (int): Interpolation order
    Returns:
        A tuple containing the transformation matrix and the rotated image.
    Note: this function is much faster applied on PIL images than on numpy ndarrays.
    """
    if isinstance(image, Image.Image):
        rows, cols = image.height, image.width
    else:
        rows, cols = image.shape[:2]
        assert len(image.shape) == 3 or len(image.shape) == 2, 'Image must be 2D or 3D'

    tform = AffineTransform(rotation=angle, scale=(1/scale, 1))
    corners = np.array([
        [0, 0],
        [0, rows - 1],
        [cols - 1, rows - 1],
        [cols - 1, 0]
    ])
    corners = tform.inverse(corners)
    minc = corners[:, 0].min()
    minr = corners[:, 1].min()
    maxc = corners[:, 0].max()
    maxr = corners[:, 1].max()
    out_rows = maxr - minr + 1
    out_cols = maxc - minc + 1
    output_shape = tuple(int(o) for o in np.around((out_rows, out_cols)))
    # fit output image in new shape
    translation = tform([[minc, minr]])
    tform = AffineTransform(rotation=angle, scale=(1/scale, 1), translation=[f for f in translation.flatten()])

    if isinstance(image, Image.Image):
        # PIL is much faster than scipy
        pdata = tform.params.flatten().tolist()[:6]
        resample = {0: Resampling.NEAREST, 1: Resampling.BILINEAR, 2: Resampling.BICUBIC, 3: Resampling.BICUBIC}.get(order, Resampling.NEAREST)
        return tform, image.transform(output_shape[::-1], Transform.AFFINE, data=pdata, resample=resample, fillcolor=cval)

    # params for scipy
    # swap X and Y axis for scipy
    pdata = tform.params.copy()[[1, 0, 2], :][:, [1, 0, 2]]
    # we copy the translation vector
    offset = pdata[:2, 2].copy()
    # scipy expects a 3x3 *linear* matrix (to include channel axis), we don't want the channel axis to be modified
    pdata[:2, 2] = 0
    if use_skimage_warp:
        return tform, warp(image, tform, output_shape=output_shape, order=order, cval=cval, clip=False, preserve_range=True)
    else:
        return tform, affine_transform(image, pdata, offset=(*offset, 0), output_shape=(*output_shape, *image.shape[2:]), cval=cval, order=order)


def line_regions(line, regions):
    """
    Filters a list of regions by line association.

    Args:
        line (list): Polyline representing the line.
        regions (list): List of region polygons

    Returns:
        A list of regions that contain the line mid-point.
    """
    mid_point = geom.LineString(line).interpolate(0.5, normalized=True)

    reg_pols = [geom.Polygon(x) for x in regions]
    regs = []
    for reg_idx, reg_pol in enumerate(reg_pols):
        if reg_pol.contains(mid_point):
            regs.append(regions[reg_idx])
    return regs


def _ray_intersect_boundaries(ray, direction, aabb):
    """
    Simplified version of [0] for 2d and AABB anchored at (0,0).

    [0] http://gamedev.stackexchange.com/questions/18436/most-efficient-aabb-vs-ray-collision-algorithms
    """
    dir_fraction = np.empty(2, dtype=ray.dtype)
    dir_fraction[direction == 0.0] = np.inf
    dir_fraction[direction != 0.0] = np.divide(1.0, direction[direction != 0.0])

    t1 = (-ray[0]) * dir_fraction[0]
    t2 = (aabb[0] - ray[0]) * dir_fraction[0]
    t3 = (-ray[1]) * dir_fraction[1]
    t4 = (aabb[1] - ray[1]) * dir_fraction[1]

    tmin = max(min(t1, t2), min(t3, t4))
    tmax = min(max(t1, t2), max(t3, t4))

    t = min(x for x in [tmin, tmax] if x >= 0)
    return ray + (direction * t)


def _calc_seam(baseline, polygon, angle, im_feats, bias=150):
    """
    Calculates seam between baseline and ROI boundary on one side.

    Adds a baseline-distance-weighted bias to the feature map, masks
    out the bounding polygon and rotates the line so it is roughly
    level.
    """
    MASK_VAL = 99999
    c_min, c_max = int(polygon[:, 0].min()), int(polygon[:, 0].max())
    r_min, r_max = int(polygon[:, 1].min()), int(polygon[:, 1].max())
    patch = im_feats[r_min:r_max+2, c_min:c_max+2].copy()
    # bias feature matrix by distance from baseline
    mask = np.ones_like(patch)
    for line_seg in zip(baseline[:-1] - (c_min, r_min), baseline[1:] - (c_min, r_min)):
        line_locs = draw.line(line_seg[0][1],
                              line_seg[0][0],
                              line_seg[1][1],
                              line_seg[1][0])
        mask[line_locs] = 0
    dist_bias = distance_transform_cdt(mask)
    # absolute mask
    mask = np.array(make_polygonal_mask(polygon-(r_min, c_min), patch.shape[::-1])) > 128
    # dilate mask to compensate for aliasing during rotation
    mask = binary_erosion(mask, border_value=True, iterations=2)
    # combine weights with features
    patch[mask] = MASK_VAL
    patch += (dist_bias*(np.mean(patch[patch != MASK_VAL])/bias))
    extrema = baseline[(0, -1), :] - (c_min, r_min)
    # scale line image to max 600 pixel width
    scale = min(1.0, 600/(c_max-c_min))
    tform, rotated_patch = _rotate(patch,
                                   angle,
                                   center=extrema[0],
                                   scale=scale,
                                   cval=MASK_VAL,
                                   use_skimage_warp=True)
    # ensure to cut off padding after rotation
    x_offsets = np.sort(np.around(tform.inverse(extrema)[:, 0]).astype('int'))
    rotated_patch = rotated_patch[:, x_offsets[0]:x_offsets[1]+1]
    # infinity pad for seamcarve
    rotated_patch = np.pad(rotated_patch, ((1, 1), (0, 0)),  mode='constant', constant_values=np.inf)
    r, c = rotated_patch.shape
    # fold into shape (c, r-2 3)
    A = np.lib.stride_tricks.as_strided(rotated_patch, (c, r-2, 3), (rotated_patch.strides[1],
                                                                     rotated_patch.strides[0],
                                                                     rotated_patch.strides[0]))
    B = rotated_patch[1:-1, 1:].swapaxes(0, 1)
    backtrack = np.zeros_like(B, dtype='int')
    T = np.empty((B.shape[1]), 'f')
    R = np.arange(-1, len(T)-1)
    for i in np.arange(c-1):
        A[i].min(1, T)
        backtrack[i] = A[i].argmin(1) + R
        B[i] += T
    # backtrack
    seam = []
    j = np.argmin(rotated_patch[1:-1, -1])
    for i in range(c-2, -2, -1):
        seam.append((i+x_offsets[0]+1, j))
        j = backtrack[i, j]
    seam = np.array(seam)[::-1]
    seam_mean = seam[:, 1].mean()
    seam_std = seam[:, 1].std()
    seam[:, 1] = np.clip(seam[:, 1], seam_mean-seam_std, seam_mean+seam_std)
    # rotate back
    seam = tform(seam).astype('int')
    # filter out seam points in masked area of original patch/in padding
    seam = seam[seam.min(axis=1) >= 0, :]
    m = (seam < mask.shape[::-1]).T
    seam = seam[np.logical_and(m[0], m[1]), :]
    seam = seam[np.invert(mask[seam.T[1], seam.T[0]])]
    seam += (c_min, r_min)
    return seam


def _extract_patch(env_up, env_bottom, baseline, offset_baseline, end_points, dir_vec, topline, offset, im_feats, bounds):
    """
    Calculate a line image patch from a ROI and the original baseline.
    """
    upper_polygon = np.concatenate((baseline, env_up[::-1]))
    bottom_polygon = np.concatenate((baseline, env_bottom[::-1]))
    upper_offset_polygon = np.concatenate((offset_baseline, env_up[::-1]))
    bottom_offset_polygon = np.concatenate((offset_baseline, env_bottom[::-1]))

    angle = np.arctan2(dir_vec[1], dir_vec[0])
    roi_polygon = unary_union([geom.Polygon(upper_polygon), geom.Polygon(bottom_polygon)])

    if topline:
        upper_seam = _calc_seam(baseline, upper_polygon, angle, im_feats)
        bottom_seam = _calc_seam(offset_baseline, bottom_offset_polygon, angle, im_feats)
    else:
        upper_seam = _calc_seam(offset_baseline, upper_offset_polygon, angle, im_feats)
        bottom_seam = _calc_seam(baseline, bottom_polygon, angle, im_feats)

    upper_seam = geom.LineString(upper_seam).simplify(5)
    bottom_seam = geom.LineString(bottom_seam).simplify(5)

    # ugly workaround against GEOM parallel_offset bug creating a
    # MultiLineString out of offset LineString
    if upper_seam.parallel_offset(offset//2, side='right').geom_type == 'MultiLineString' or offset == 0:
        upper_seam = np.array(upper_seam.coords, dtype=int)
    else:
        upper_seam = np.array(upper_seam.parallel_offset(offset//2, side='right').coords, dtype=int)[::-1]
    if bottom_seam.parallel_offset(offset//2, side='left').geom_type == 'MultiLineString' or offset == 0:
        bottom_seam = np.array(bottom_seam.coords, dtype=int)
    else:
        bottom_seam = np.array(bottom_seam.parallel_offset(offset//2, side='left').coords, dtype=int)

    # offsetting might produce bounds outside the image. Clip it to the image bounds.
    polygon = np.concatenate(([end_points[0]], upper_seam, [end_points[-1]], bottom_seam[::-1]))
    polygon = geom.Polygon(polygon)
    if not polygon.is_valid:
        polygon = np.concatenate(([end_points[-1]], upper_seam, [end_points[0]], bottom_seam))
        polygon = geom.Polygon(polygon)
    if not polygon.is_valid:
        raise Exception(f'Invalid bounding polygon computed: {explain_validity(polygon)}')
    polygon = np.array(roi_polygon.intersection(polygon).boundary.coords, dtype=int)
    return polygon


def _calc_roi(line, bounds, baselines, suppl_obj, p_dir):
    # interpolate baseline
    ls = geom.LineString(line)
    ip_line = [line[0]]
    dist = 10
    while dist < ls.length:
        ip_line.append(np.array(ls.interpolate(dist).coords[0]))
        dist += 10
    ip_line.append(line[-1])
    ip_line = np.array(ip_line)
    upper_bounds_intersects = []
    bottom_bounds_intersects = []
    for point in ip_line:
        upper_bounds_intersects.append(_ray_intersect_boundaries(point, (p_dir*(-1, 1))[::-1], bounds+1).astype('int'))
        bottom_bounds_intersects.append(_ray_intersect_boundaries(point, (p_dir*(1, -1))[::-1], bounds+1).astype('int'))
    # build polygon between baseline and bbox intersects
    upper_polygon = geom.Polygon(ip_line.tolist() + upper_bounds_intersects)
    bottom_polygon = geom.Polygon(ip_line.tolist() + bottom_bounds_intersects)

    # select baselines at least partially in each polygon
    side_a = [geom.LineString(upper_bounds_intersects)]
    side_b = [geom.LineString(bottom_bounds_intersects)]

    for adj_line in baselines + suppl_obj:
        adj_line = geom.LineString(adj_line)
        if upper_polygon.intersects(adj_line):
            side_a.append(adj_line)
        elif bottom_polygon.intersects(adj_line):
            side_b.append(adj_line)
    side_a = unary_union(side_a).buffer(1).boundary
    side_b = unary_union(side_b).buffer(1).boundary

    def _find_closest_point(pt, intersects):
        spt = geom.Point(pt)
        if intersects.is_empty:
            raise Exception(f'No intersection with boundaries. Shapely intersection object: {intersects.wkt}')
        if intersects.geom_type == 'MultiPoint':
            return min([p for p in intersects.geoms], key=lambda x: spt.distance(x))
        elif intersects.geom_type == 'Point':
            return intersects
        elif intersects.geom_type == 'GeometryCollection' and len(intersects.geoms) > 0:
            t = min([p for p in intersects.geoms], key=lambda x: spt.distance(x))
            if t == 'Point':
                return t
            else:
                return nearest_points(spt, t)[1]
        else:
            raise Exception(f'No intersection with boundaries. Shapely intersection object: {intersects.wkt}')

    env_up = []
    env_bottom = []
    # find orthogonal (to linear regression) intersects with adjacent objects to complete roi
    for point, upper_bounds_intersect, bottom_bounds_intersect in zip(ip_line, upper_bounds_intersects, bottom_bounds_intersects):
        upper_limit = _find_closest_point(point, geom.LineString(
            [point, upper_bounds_intersect]).intersection(side_a))
        bottom_limit = _find_closest_point(point, geom.LineString(
            [point, bottom_bounds_intersect]).intersection(side_b))
        env_up.append(upper_limit.coords[0])
        env_bottom.append(bottom_limit.coords[0])
    env_up = np.array(env_up, dtype='uint')
    env_bottom = np.array(env_bottom, dtype='uint')
    return env_up, env_bottom


def calculate_polygonal_environment(im: Image.Image = None,
                                    baselines: Sequence[Sequence[Tuple[int, int]]] = None,
                                    suppl_obj: Sequence[Sequence[Tuple[int, int]]] = None,
                                    im_feats: np.ndarray = None,
                                    scale: Tuple[int, int] = None,
                                    topline: bool = False,
                                    raise_on_error: bool = False):
    """
    Given a list of baselines and an input image, calculates a polygonal
    environment around each baseline.

    Args:
        im: grayscale input image (mode 'L')
        baselines: List of lists containing a single baseline per entry.
        suppl_obj: List of lists containing additional polylines that should be
                   considered hard boundaries for polygonizaton purposes. Can
                   be used to prevent polygonization into non-text areas such
                   as illustrations or to compute the polygonization of a
                   subset of the lines in an image.
        im_feats: An optional precomputed seamcarve energy map. Overrides data
                  in `im`. The default map is `gaussian_filter(sobel(im), 2)`.
        scale: A 2-tuple (h, w) containing optional scale factors of the input.
               Values of 0 are used for aspect-preserving scaling. `None` skips
               input scaling.
        topline: Switch to change default baseline location for offset
                 calculation purposes. If set to False, baselines are assumed
                 to be on the bottom of the text line and will be offset
                 upwards, if set to True, baselines are on the top and will be
                 offset downwards. If set to None, no offset will be applied.
        raise_on_error: Raises error instead of logging them when they are
                        not-blocking
    Returns:
        List of lists of coordinates. If no polygonization could be compute for
        a baseline `None` is returned instead.
    """
    if scale is not None and (scale[0] > 0 or scale[1] > 0):
        w, h = im.size
        oh, ow = scale
        if oh == 0:
            oh = int(h * ow/w)
        elif ow == 0:
            ow = int(w * oh/h)
        im = im.resize((ow, oh))
        scale = np.array((ow/w, oh/h))
        # rescale baselines
        baselines = [(np.array(bl) * scale).astype('int').tolist() for bl in baselines]
        # rescale suppl_obj
        if suppl_obj is not None:
            suppl_obj = [(np.array(bl) * scale).astype('int').tolist() for bl in suppl_obj]

    if im_feats is None:
        bounds = np.array(im.size, dtype=float) - 1
        im = np.array(im.convert('L'))
        # compute image gradient
        im_feats = gaussian_filter(sobel(im), 0.5)
    else:
        bounds = np.array(im_feats.shape[::-1], dtype=float) - 1

    polygons = []
    if suppl_obj is None:
        suppl_obj = []

    for idx, line in enumerate(baselines):
        try:
            end_points = (line[0], line[-1])
            line = geom.LineString(line)
            offset = default_specs.SEGMENTATION_HYPER_PARAMS['line_width'] if topline is not None else 0
            offset_line = line.parallel_offset(offset, side='left' if topline else 'right')
            line = np.array(line.coords, dtype=float)
            offset_line = np.array(offset_line.coords, dtype=float)

            # parallel_offset on the right reverses the coordinate order
            if not topline:
                offset_line = offset_line[::-1]
            # calculate magnitude-weighted average direction vector
            lengths = np.linalg.norm(np.diff(line.T), axis=0)
            p_dir = np.mean(np.diff(line.T) * lengths/lengths.sum(), axis=1)
            p_dir = (p_dir.T / np.sqrt(np.sum(p_dir**2, axis=-1)))

            env_up, env_bottom = _calc_roi(line, bounds, baselines[:idx] + baselines[idx+1:], suppl_obj, p_dir)

            polygons.append(_extract_patch(env_up,
                                           env_bottom,
                                           line.astype('int'),
                                           offset_line.astype('int'),
                                           end_points,
                                           p_dir,
                                           topline,
                                           offset,
                                           im_feats,
                                           bounds))
        except Exception as e:
            if raise_on_error:
                raise
            logger.warning(f'Polygonizer failed on line {idx}: {e}')
            polygons.append(None)

    if scale is not None:
        polygons = [(np.array(pol)/scale).astype('uint').tolist() if pol is not None else None for pol in polygons]
    return polygons


def polygonal_reading_order(lines: Sequence[Dict],
                            text_direction: Literal['lr', 'rl'] = 'lr',
                            regions: Optional[Sequence[geom.Polygon]] = None) -> Sequence[int]:
    """
    Given a list of baselines and regions, calculates the correct reading order
    and applies it to the input.

    Args:
        lines: List of tuples containing the baseline and its polygonization.
        regions: List of region polygons.
        text_direction: Set principal text direction for column ordering. Can
                        be 'lr' or 'rl'

    Returns:
        The indices of the ordered input.
    """
    lines = [(line['tags']['type'], line['baseline'], line['boundary']) for line in lines]

    bounds = []
    if regions is None:
        regions = []
    region_lines = [[] for _ in range(len(regions))]
    indizes = {}
    for line_idx, line in enumerate(lines):
        s_line = geom.LineString(line[1])
        in_region = False
        for idx, reg in enumerate(regions):
            if is_in_region(s_line, reg):
                region_lines[idx].append((line_idx, (slice(s_line.bounds[1], s_line.bounds[3]),
                                                     slice(s_line.bounds[0], s_line.bounds[2]))))
                in_region = True
                break
        if not in_region:
            bounds.append((slice(s_line.bounds[1], s_line.bounds[3]),
                           slice(s_line.bounds[0], s_line.bounds[2])))
            indizes[line_idx] = ('line', line)
    # order everything in regions
    intra_region_order = [[] for _ in range(len(regions))]
    for idx, reg in enumerate(regions):
        if len(region_lines[idx]) > 0:
            order = reading_order([x[1] for x in region_lines[idx]], text_direction)
            lsort = topsort(order)
            intra_region_order[idx] = [region_lines[idx][i][0] for i in lsort]
            reg = reg.bounds
            bounds.append((slice(reg[1], reg[3]), slice(reg[0], reg[2])))
            indizes[line_idx+idx+1] = ('region', idx)
    # order unassigned lines and regions
    order = reading_order(bounds, text_direction)
    lsort = topsort(order)
    sidz = sorted(indizes.keys())
    lsort = [sidz[i] for i in lsort]
    ordered_idxs = []
    for i in lsort:
        if indizes[i][0] == 'line':
            ordered_idxs.append(i)
        else:
            ordered_idxs.extend(intra_region_order[indizes[i][1]])
    return ordered_idxs


def is_in_region(line: geom.LineString, region: geom.Polygon) -> bool:
    """
    Tests if a line is inside a region, i.e. if the mid point of the baseline
    is inside the region.

    Args:
        line: line to test
        region: region to test against

    Returns:
        False if line is not inside region, True otherwise.
    """
    l_obj = line.interpolate(0.5, normalized=True)
    return region.contains(l_obj)


def neural_reading_order(lines: Sequence[Dict],
                         text_direction: str = 'lr',
                         regions: Optional[Sequence[geom.Polygon]] = None,
                         im_size: Tuple[int, int] = None,
                         model: 'TorchVGSLModel' = None,
                         class_mapping: Dict[str, int] = None) -> Sequence[int]:
    """
    Given a list of baselines and regions, calculates the correct reading order
    and applies it to the input.

    Args:
        lines: List of tuples containing the baseline and its polygonization.
        model: torch Module for

    Returns:
        The indices of the ordered input.
    """
    lines = [(line['tags']['type'], line['baseline'], line['boundary']) for line in lines]
    # construct all possible pairs
    h, w = im_size
    features = []
    for i in lines:
        for j in lines:
            if i == j and len(lines) != 1:
                continue
            num_classes = len(class_mapping) + 1
            cl_i = torch.zeros(num_classes, dtype=torch.float)
            cl_j = torch.zeros(num_classes, dtype=torch.float)
            cl_i[class_mapping.get(i[0], 0)] = 1
            cl_j[class_mapping.get(j[0], 0)] = 1
            line_coords_i = np.array(i[1]) / (w, h)
            line_center_i = np.mean(line_coords_i, axis=0)
            line_coords_j = np.array(j[1]) / (w, h)
            line_center_j = np.mean(line_coords_j, axis=0)
            features.append(torch.cat((cl_i,
                                       torch.tensor(line_center_i, dtype=torch.float),  # lin
                                       torch.tensor(line_coords_i[0, :], dtype=torch.float),
                                       torch.tensor(line_coords_i[-1, :], dtype=torch.float),
                                       cl_j,
                                       torch.tensor(line_center_j, dtype=torch.float),  # lin
                                       torch.tensor(line_coords_j[0, :], dtype=torch.float),
                                       torch.tensor(line_coords_j[-1, :], dtype=torch.float))))
    features = torch.stack(features)
    output = F.sigmoid(model(features))

    order = torch.zeros((len(lines), len(lines)))
    idx = 0
    for i in range(len(lines)):
        for j in range(len(lines)):
            if i == j and len(lines) != 1:
                continue
            order[i, j] = output[idx]
            idx += 1
    # decode order relation matrix
    path = _greedy_order_decoder(order)
    return path


def _greedy_order_decoder(P):
    """
    A greedy decoder of order-relation matrix. For each position in the
    reading order we select the most probable one, then move to the next
    position. Most probable for position:

    .. math::
        z^{\\star}_t = \\argmax_{(s,\\nu) \\ni z^{\\star}}
        \\prod_{(s',\\nu') \\in z^\\star}{\\tilde{P}(Y=1\\mid s',s)}
        \\times \\prod_{\\substack{(s'',\\nu'') \\ni z^\\star\\
         s'' \\ne s}}{\\tilde{P}(r=0\\mid s'',s)}, 1\\le t \\le n
    """
    A = P + torch.finfo(torch.float).eps
    N = P.shape[0]
    A = (A + (1-A).T)/2
    for i in range(A.shape[0]):
        A[i, i] = torch.finfo(torch.float).eps
    best_path = []
    # use log(p(R\mid s',s)) to shift multiplication to sum
    lP = torch.log(A)
    for i in range(N):
        lP[i, i] = 0
    for t in range(N):
        for i in range(N):
            idx = torch.argmax(lP.sum(axis=1))
            if idx not in best_path:
                best_path.append(idx)
                lP[idx, :] = lP[:, idx]
                lP[:, idx] = 0
                break
    return torch.tensor(best_path)


def scale_regions(regions: Sequence[Tuple[List[int], List[int]]],
                  scale: Union[float, Tuple[float, float]]) -> Sequence[Tuple[List, List]]:
    """
    Scales baselines/polygon coordinates by a certain factor.

    Args:
        lines: List of tuples containing the baseline and its polygonization.
        scale: Scaling factor
    """
    if isinstance(scale, float):
        scale = (scale, scale)
    scaled_regions = []
    for region in regions:
        scaled_regions.append((np.array(region) * scale).astype('uint').tolist())
    return scaled_regions


def scale_polygonal_lines(lines: Sequence[Tuple[List, List]], scale: Union[float, Tuple[float, float]]) -> Sequence[Tuple[List, List]]:
    """
    Scales baselines/polygon coordinates by a certain factor.

    Args:
        lines: List of tuples containing the baseline and its polygonization.
        scale: Scaling factor
    """
    if isinstance(scale, float):
        scale = (scale, scale)
    scaled_lines = []
    for line in lines:
        bl, pl = line
        scaled_lines.append(((np.array(bl) * scale).astype('int').tolist(),
                             (np.array(pl) * scale).astype('int').tolist()))
    return scaled_lines


def _test_intersect(bp, uv, bs):
    """
    Returns the intersection points of a ray with direction `uv` from
    `bp` with a polygon `bs`.
    """
    u = bp - np.roll(bs, 2)
    v = bs - np.roll(bs, 2)
    points = []
    for dir in ((1, -1), (-1, 1)):
        w = (uv * dir * (1, -1))[::-1]
        z = np.dot(v, w)
        t1 = np.cross(v, u) / (z + np.finfo(float).eps)
        t2 = np.dot(u, w) / (z + np.finfo(float).eps)
        t1 = t1[np.logical_and(t2 >= 0.0, t2 <= 1.0)]
        points.extend(bp + (t1[np.where(t1 >= 0)[0].min()] * (uv * dir)))
    return np.array(points)


def compute_polygon_section(baseline: Sequence[Tuple[int, int]],
                            boundary: Sequence[Tuple[int, int]],
                            dist1: int,
                            dist2: int) -> Tuple[Tuple[int, int]]:
    """
    Given a baseline, polygonal boundary, and two points on the baseline return
    the rectangle formed by the orthogonal cuts on that baseline segment. The
    resulting polygon is not garantueed to have a non-zero area.

    The distance can be larger than the actual length of the baseline if the
    baseline endpoints are inside the bounding polygon. In that case the
    baseline will be extrapolated to the polygon edge.

    Args:
        baseline: A polyline ((x1, y1), ..., (xn, yn))
        boundary: A bounding polygon around the baseline (same format as
                  baseline). Last and first point are automatically connected.
        dist1: Absolute distance along the baseline of the first point.
        dist2: Absolute distance along the baseline of the second point.

    Returns:
        A sequence of polygon points.
    """
    # find baseline segments the points are in
    if dist1 == 0:
        dist1 = np.finfo(float).eps
    if dist2 == 0:
        dist2 = np.finfo(float).eps
    boundary_pol = geom.Polygon(boundary)
    bl = np.array(baseline)
    # extend first/last segment of baseline if not on polygon boundary
    if boundary_pol.contains(geom.Point(bl[0])):
        logger.debug(f'Extending leftmost end of baseline {bl} to polygon boundary')
        l_point = boundary_pol.boundary.intersection(geom.LineString(
            [(bl[0][0]-10*(bl[1][0]-bl[0][0]), bl[0][1]-10*(bl[1][1]-bl[0][1])), bl[0]]))
        # intersection is incidental with boundary so take closest point instead
        if l_point.geom_type != 'Point':
            bl[0] = np.array(nearest_points(geom.Point(bl[0]), boundary_pol)[1].coords[0], 'int')
        else:
            bl[0] = np.array(l_point.coords[0], 'int')
    if boundary_pol.contains(geom.Point(bl[-1])):
        logger.debug(f'Extending rightmost end of baseline {bl} to polygon boundary')
        r_point = boundary_pol.boundary.intersection(geom.LineString(
            [(bl[-1][0]-10*(bl[-2][0]-bl[-1][0]), bl[-1][1]-10*(bl[-2][1]-bl[-1][1])), bl[-1]]))
        if r_point.geom_type != 'Point':
            bl[-1] = np.array(nearest_points(geom.Point(bl[-1]), boundary_pol)[1].coords[0], 'int')
        else:
            bl[-1] = np.array(r_point.coords[0], 'int')
    dist1 = min(geom.LineString(bl).length - np.finfo(float).eps, dist1)
    dist2 = min(geom.LineString(bl).length - np.finfo(float).eps, dist2)
    dists = np.cumsum(np.diag(np.roll(squareform(pdist(bl)), 1)))
    segs_idx = np.searchsorted(dists, [dist1, dist2])
    segs = np.dstack((bl[segs_idx-1], bl[segs_idx]))
    # compute unit vector of segments (NOT orthogonal)
    norm_vec = (segs[..., 1] - segs[..., 0])
    norm_vec_len = np.sqrt(np.sum(norm_vec**2, axis=1))
    unit_vec = norm_vec / np.tile(norm_vec_len, (2, 1)).T
    # find point start/end point on segments
    seg_dists = (dist1, dist2) - dists[segs_idx-1]
    seg_points = segs[..., 0] + (seg_dists * unit_vec.T).T
    # get intersects
    bounds = np.array(boundary)
    try:
        points = [_test_intersect(point, uv[::-1], bounds).round() for point, uv in zip(seg_points, unit_vec)]
    except ValueError:
        logger.debug('No intercepts with polygon (possibly misshaped polygon)')
        return seg_points.astype('int').tolist()
    o = np.int_(points[0]).reshape(-1, 2).tolist()
    o.extend(np.int_(np.roll(points[1], 2)).reshape(-1, 2).tolist())
    return tuple(o)


def _bevelled_warping_envelope(baseline: np.ndarray,
                               output_bl_start: Tuple[float, float],
                               output_shape: Tuple[int, int]) -> Tuple[List[Tuple[int, int]], List[Tuple[int, int]]]:
    """
    Calculates the source and target envelope for a piecewise affine transform
    """
    def _as_int_tuple(x):
        return tuple(int(i) for i in x)

    envelope_dy = [-output_bl_start[1], output_shape[0] - output_bl_start[1]]
    diff_bl = np.diff(baseline, axis=0)
    diff_bl_normed = diff_bl / np.linalg.norm(diff_bl, axis=1)[:, None]
    l_bl = len(baseline)
    cum_lens = np.cumsum([0] + np.linalg.norm(diff_bl, axis=1).tolist())

    bl_seg_normals = np.array([-diff_bl_normed[:, 1], diff_bl_normed[:, 0]]).T
    ini_point = baseline[0] - diff_bl_normed[0] * output_bl_start[0]
    source_envelope = [
        _as_int_tuple(ini_point + envelope_dy[0]*bl_seg_normals[0]),
        _as_int_tuple(ini_point + envelope_dy[1]*bl_seg_normals[0]),
    ]
    target_envelope = [
        (0, 0),
        (0, output_shape[0])
    ]
    MAX_BEVEL_WIDTH = output_shape[0] / 3
    BEVEL_STEP_WIDTH = MAX_BEVEL_WIDTH / 2

    for k in range(l_bl-2):
        pt = baseline[k+1]
        seg_prev = baseline[k] - pt
        seg_next = baseline[k+2] - pt
        bevel_prev = seg_prev / max(2., np.linalg.norm(seg_prev) / MAX_BEVEL_WIDTH)
        bevel_next = seg_next / max(2., np.linalg.norm(seg_next) / MAX_BEVEL_WIDTH)
        bevel_nsteps = max(1, np.round((np.linalg.norm(bevel_prev) + np.linalg.norm(bevel_next)) / BEVEL_STEP_WIDTH))
        l_prev = np.linalg.norm(bevel_prev)
        l_next = np.linalg.norm(bevel_next)
        for i in range(int(bevel_nsteps)+1):
            # bezier interp
            t = i / bevel_nsteps
            tpt = pt + (1-t)**2 * bevel_prev + t**2 * bevel_next
            tx = output_bl_start[0] + cum_lens[k+1] - (1-t)**2 * l_prev + t**2 * l_next
            tnormal = (1-t) * bl_seg_normals[k] + t * bl_seg_normals[k+1]
            tnormal /= np.linalg.norm(tnormal)
            source_points = [_as_int_tuple(tpt + envelope_dy[0]*tnormal), _as_int_tuple(tpt + envelope_dy[1]*tnormal)]
            target_points = [(int(tx), 0), (int(tx), output_shape[0])]
            # avoid duplicate points leading to singularities
            if source_points[0] == source_envelope[-2] or source_points[1] == source_envelope[-1] or target_points[0] == target_envelope[-2]:
                continue
            source_envelope += source_points
            target_envelope += target_points

    end_point = baseline[-1] + diff_bl_normed[-1]*(output_shape[1]-cum_lens[-1]-output_bl_start[0])
    source_envelope += [
        end_point + envelope_dy[0]*bl_seg_normals[-1],
        end_point + envelope_dy[1]*bl_seg_normals[-1],
    ]
    target_envelope += [
        (output_shape[1], 0),
        (output_shape[1], output_shape[0])
    ]
    return source_envelope, target_envelope


def make_polygonal_mask(polygon: np.ndarray, shape: Tuple[int, int]) -> Image.Image:
    """
    Creates a mask from a polygon.

    Args:
        polygon: A polygon as a list of points.
        shape: The shape of the mask to create.

    Returns:
        A PIL.Image.Image instance containing the mask.
    """
    mask = Image.new('L', shape, 0)
    ImageDraw.Draw(mask).polygon([tuple(p) for p in polygon.astype(int).tolist()], fill=255, width=2)
    return mask


def apply_polygonal_mask(img: Image.Image, polygon: np.ndarray, cval: int = 0) -> Image.Image:
    """
    Extract the polygonal mask of an image.
    """
    mask = make_polygonal_mask(polygon, img.size)
    out = Image.new(img.mode, (img.width, img.height), cval)
    out.paste(img, mask=mask)
    return out


def extract_polygons(im: Image.Image,
                     bounds: "Segmentation",
                     legacy: bool = False) -> Generator[Tuple[Image.Image, Union["BBoxLine", "BaselineLine"],], None, None]:
    """
    Yields the subimages of image im defined in the list of bounding polygons
    with baselines preserving order.

    Args:
        im: Input image
        bounds: A Segmentation class containing a bounding box or baseline
                segmentation.
        legacy: Use the old, slow, and deprecated path

    Yields:
        The extracted subimage, and the corresponding bounding box or baseline
    """
    if bounds.type == 'baselines':
        # select proper interpolation scheme depending on shape
        if im.mode == '1':
            order = 0
            im = im.convert('L')
        else:
            order = 1

        for line in bounds.lines:
            if line.boundary is None:
                raise KrakenInputException('No boundary given for line')
            if len(line.baseline) < 2 or geom.LineString(line.baseline).length < 5:
                raise KrakenInputException('Baseline length below minimum 5px')
            pl = np.array(line.boundary)
            baseline = np.array(line.baseline)
            c_min, c_max = int(pl[:, 0].min()), int(pl[:, 0].max())
            r_min, r_max = int(pl[:, 1].min()), int(pl[:, 1].max())

            imshape = np.array([im.height, im.width])

            if (pl < 0).any() or (pl.max(axis=0)[::-1] >= imshape).any():
                raise KrakenInputException('Line polygon outside of image bounds')
            if (baseline < 0).any() or (baseline.max(axis=0)[::-1] >= imshape).any():
                raise KrakenInputException('Baseline outside of image bounds')

            if legacy:
                im = np.asarray(im)
                # Old, slow, and deprecated path
                # fast path for straight baselines requiring only rotation
                if len(baseline) == 2:
                    baseline = baseline.astype(float)
                    # calculate direction vector
                    lengths = np.linalg.norm(np.diff(baseline.T), axis=0)
                    p_dir = np.mean(np.diff(baseline.T) * lengths/lengths.sum(), axis=1)
                    p_dir = (p_dir.T / np.sqrt(np.sum(p_dir**2, axis=-1)))
                    angle = np.arctan2(p_dir[1], p_dir[0])
                    patch = im[r_min:r_max+1, c_min:c_max+1].copy()
                    offset_polygon = pl - (c_min, r_min)
                    offset_polygon2 = offset_polygon.flatten().tolist()
                    img = Image.new('L', patch.shape[:2][::-1], 0)
                    ImageDraw.Draw(img).polygon(offset_polygon2, outline=1, fill=1)
                    mask = np.asarray(img, dtype=bool)
                    patch[np.invert(mask)] = 0
                    extrema = offset_polygon[(0, -1), :]
                    # scale line image to max 600 pixel width
                    tform, rotated_patch = _rotate(patch, angle, center=extrema[0], scale=1.0, cval=0)
                    i = Image.fromarray(rotated_patch.astype('uint8'))
                # normal slow path with piecewise affine transformation
                else:
                    if len(pl) > 50:
                        pl = approximate_polygon(pl, 2)
                    full_polygon = subdivide_polygon(pl, preserve_ends=True)
                    pl = geom.MultiPoint(full_polygon)

                    bl = zip(baseline[:-1:], baseline[1::])
                    bl = [geom.LineString(x) for x in bl]
                    cum_lens = np.cumsum([0] + [line.length for line in bl])
                    # distance of intercept from start point and number of line segment
                    control_pts = []
                    for point in pl.geoms:
                        npoint = np.array(point.coords)[0]
                        line_idx, dist, intercept = min(((idx, line.project(point),
                                                        np.array(line.interpolate(line.project(point)).coords)) for idx, line in enumerate(bl)),
                                                        key=lambda x: np.linalg.norm(npoint-x[2]))
                        # absolute distance from start of line
                        line_dist = cum_lens[line_idx] + dist
                        intercept = np.array(intercept)
                        # side of line the point is at
                        side = np.linalg.det(np.array([[baseline[line_idx+1][0]-baseline[line_idx][0],
                                                        npoint[0]-baseline[line_idx][0]],
                                                       [baseline[line_idx+1][1]-baseline[line_idx][1],
                                                           npoint[1]-baseline[line_idx][1]]]))
                        side = np.sign(side)
                        # signed perpendicular distance from the rectified distance
                        per_dist = side * np.linalg.norm(npoint-intercept)
                        control_pts.append((line_dist, per_dist))
                    # calculate baseline destination points
                    bl_dst_pts = baseline[0] + np.dstack((cum_lens, np.zeros_like(cum_lens)))[0]
                    # calculate bounding polygon destination points
                    pol_dst_pts = np.array([baseline[0] + (line_dist, per_dist) for line_dist, per_dist in control_pts])
                    # extract bounding box patch
                    c_dst_min, c_dst_max = int(pol_dst_pts[:, 0].min()), int(pol_dst_pts[:, 0].max())
                    r_dst_min, r_dst_max = int(pol_dst_pts[:, 1].min()), int(pol_dst_pts[:, 1].max())
                    output_shape = np.around((r_dst_max - r_dst_min + 1, c_dst_max - c_dst_min + 1))
                    patch = im[r_min:r_max+1, c_min:c_max+1].copy()
                    # offset src points by patch shape
                    offset_polygon = full_polygon - (c_min, r_min)
                    offset_baseline = baseline - (c_min, r_min)
                    # offset dst point by dst polygon shape
                    offset_bl_dst_pts = bl_dst_pts - (c_dst_min, r_dst_min)
                    offset_pol_dst_pts = pol_dst_pts - (c_dst_min, r_dst_min)
                    # mask out points outside bounding polygon
                    offset_polygon2 = offset_polygon.flatten().tolist()
                    img = Image.new('L', patch.shape[:2][::-1], 0)
                    ImageDraw.Draw(img).polygon(offset_polygon2, outline=1, fill=1)
                    mask = np.asarray(img, dtype=bool)
                    patch[np.invert(mask)] = 0
                    # estimate piecewise transform
                    src_points = np.concatenate((offset_baseline, offset_polygon))
                    dst_points = np.concatenate((offset_bl_dst_pts, offset_pol_dst_pts))
                    tform = FastPiecewiseAffineTransform()
                    tform.estimate(src_points, dst_points)
                    o = warp(patch, tform.inverse, output_shape=output_shape, preserve_range=True, order=order)
                    i = Image.fromarray(o.astype('uint8'))

            else:  # if not legacy
                # new, fast, and efficient path
                # fast path for straight baselines requiring only rotation
                if len(baseline) == 2:
                    baseline = baseline.astype(float)
                    # calculate direction vector
                    lengths = np.linalg.norm(np.diff(baseline.T), axis=0)
                    p_dir = np.mean(np.diff(baseline.T) * lengths/lengths.sum(), axis=1)
                    p_dir = (p_dir.T / np.sqrt(np.sum(p_dir**2, axis=-1)))
                    angle = np.arctan2(p_dir[1], p_dir[0])
                    # crop out bounding box
                    patch = im.crop((c_min, r_min, c_max+1, r_max+1))
                    offset_polygon = pl - (c_min, r_min)
                    patch = apply_polygonal_mask(patch, offset_polygon, cval=0)
                    extrema = offset_polygon[(0, -1), :]
                    tform, i = _rotate(patch, angle, center=extrema[0], scale=1.0, cval=0, order=order)
                # normal slow path with piecewise affine transformation
                else:
                    if len(pl) > 50:
                        pl = approximate_polygon(pl, 2)
                    full_polygon = subdivide_polygon(pl, preserve_ends=True)

                    # baseline segment vectors
                    diff_bl = np.diff(baseline, axis=0)
                    diff_bl_norms = np.linalg.norm(diff_bl, axis=1)
                    diff_bl_normed = diff_bl / diff_bl_norms[:, None]

                    l_poly = len(full_polygon)
                    cum_lens = np.cumsum([0] + np.linalg.norm(diff_bl, axis=1).tolist())

                    # calculate baseline destination points :
                    bl_dst_pts = baseline[0] + np.dstack((cum_lens, np.zeros_like(cum_lens)))[0]

                    # calculate bounding polygon destination points :
                    # diff[k, p] = baseline[k] - polygon[p]
                    poly_bl_diff = full_polygon[None, :] - baseline[:-1, None]
                    # local x coordinates of polygon points on baseline segments
                    # x[k, p] = (baseline[k] - polygon[p]) . (baseline[k+1] - baseline[k]) / |baseline[k+1] - baseline[k]|
                    poly_bl_x = np.einsum('kpm,km->kp', poly_bl_diff, diff_bl_normed)
                    # distance to baseline segments
                    poly_bl_segdist = np.maximum(-poly_bl_x, poly_bl_x - diff_bl_norms[:, None])
                    # closest baseline segment index
                    poly_closest_bl = np.argmin((poly_bl_segdist), axis=0)
                    poly_bl_x = poly_bl_x[poly_closest_bl, np.arange(l_poly)]
                    poly_bl_diff = poly_bl_diff[poly_closest_bl, np.arange(l_poly)]
                    # signed distance between polygon points and baseline segments (to get y coordinates)
                    poly_bl_y = np.cross(diff_bl_normed[poly_closest_bl], poly_bl_diff)
                    # final destination points
                    pol_dst_pts = np.array(
                        [cum_lens[poly_closest_bl] + poly_bl_x, poly_bl_y]
                    ).T + baseline[:1]

                    # extract bounding box patch
                    c_dst_min, c_dst_max = int(pol_dst_pts[:, 0].min()), int(pol_dst_pts[:, 0].max())
                    r_dst_min, r_dst_max = int(pol_dst_pts[:, 1].min()), int(pol_dst_pts[:, 1].max())
                    output_shape = np.around((r_dst_max - r_dst_min + 1, c_dst_max - c_dst_min + 1))
                    patch = im.crop((c_min, r_min, c_max+1, r_max+1))
                    # offset src points by patch shape
                    offset_polygon = full_polygon - (c_min, r_min)
                    offset_baseline = baseline - (c_min, r_min)
                    # offset dst point by dst polygon shape
                    offset_bl_dst_pts = bl_dst_pts - (c_dst_min, r_dst_min)
                    # mask out points outside bounding polygon
                    patch = apply_polygonal_mask(patch, offset_polygon, cval=0)

                    # estimate piecewise transform by beveling angles
                    source_envelope, target_envelope = _bevelled_warping_envelope(offset_baseline, offset_bl_dst_pts[0], output_shape)
                    # mesh for PIL, as (box, quad) tuples : box is (NW, SE) and quad is (NW, SW, SE, NE)
                    deform_mesh = [
                        (
                            (*target_envelope[i], *target_envelope[i+3]),
                            (*source_envelope[i], *source_envelope[i+1], *source_envelope[i+3], *source_envelope[i+2])
                        )
                        for i in range(0, len(source_envelope)-3, 2)
                    ]
                    # warp
                    resample = {0: Resampling.NEAREST, 1: Resampling.BILINEAR, 2: Resampling.BICUBIC, 3: Resampling.BICUBIC}.get(order, Resampling.NEAREST)
                    i = patch.transform((output_shape[1], output_shape[0]), Image.MESH, data=deform_mesh, resample=resample)

            yield i.crop(i.getbbox()), line
    else:
        if bounds.text_direction.startswith('vertical'):
            angle = 90
        else:
            angle = 0
        for line in bounds.lines:
            box = line.bbox
            if isinstance(box, tuple):
                box = list(box)
            if (box < [0, 0, 0, 0] or box[::2] >= [im.size[0], im.size[0]] or
                    box[1::2] >= [im.size[1], im.size[1]]):
                logger.error('bbox {} is outside of image bounds {}'.format(box, im.size))
                raise KrakenInputException('Line outside of image bounds')
            yield im.crop(box).rotate(angle, expand=True), box
